import pyomo.environ as pyo
import pyomo.gdp as gdp
import copy
import numpy as np
import statsmodels.api as sm
import pandas as pd

from ..utilities import fit_piecewise_function, fit_linear_function
from ..technology import Technology
from ...utilities import link_full_resolution_to_clustered

import logging

log = logging.getLogger(__name__)


class HeatPump(Technology):
    """
    Heat pump

    Resembles a heat pump
    Three different types of heat pumps are possible: air sourced (
    'HeatPump_AirSourced'), ground sourced ('HeatPump_GroundSourced') and water
    sourced ('HeatPump_WaterSourced'). Additionally, a heating curve is determined for
    heating for buildings. Then, the application needs to be set to either
    'floor_heating' or 'radiator_heating' in the data file. Otherwise, the output
    temperature of the heat pump can also be set to a given temperature. The
    coefficient of performance at full load is calculated in the respective fitting
    function with the equations provided in Ruhnau, O., Hirth, L., & Praktiknjo,
    A. (2019). Time series of heat demand and heat pump efficiency for energy system
    modeling. Scientific Data, 6(1). https://doi.org/10.1038/s41597-019-0199-y

    The part load behavior is modelled after equation (3) in Xu, Z., Li, H., Xu, W.,
    Shao, S., Wang, Z., Gou, X., Zhao, M., & Li, J. (2022). Investigation on the
    efficiency degradation characterization of low ambient temperature air source
    heat pump under partial load operation. International Journal of Refrigeration,
    133, 99–110. https://doi.org/10.1016/J.IJREFRIG.2021.10.002

    Essentially, the equations for the heat pump model are the same as for generic
    conversion technology type 1 (with time-dependent performance parameter).

    Ramping rates of the technology can be constraint.

    .. math::
      -rampingrate \\leq \\sum(Input_{t, car}) - \\sum(Input_{t-1, car}) \\leq rampingrate

    """

    def __init__(self, tec_data: dict):
        """
        Constructor

        :param dict tec_data: technology data
        """
        super().__init__(tec_data)

        self.component_options.emissions_based_on = "input"
        self.component_options.main_input_carrier = tec_data["Performance"][
            "main_input_carrier"
        ]

    def fit_technology_performance(self, climate_data: pd.DataFrame, location: dict):
        """
        Performs fitting for technology type HeatPump

        :param tec_data: technology data
        :param climate_data: climate data
        :return:
        """
        super(HeatPump, self).fit_technology_performance(climate_data, location)

        # Climate data & Number of timesteps
        time_steps = len(climate_data)

        # Ambient air temperature
        T = copy.deepcopy(climate_data["temp_air"])

        # Determine T_out
        if self.input_parameters.performance_data["application"] == "radiator_heating":
            t_out = 40 - T
        elif self.input_parameters.performance_data["application"] == "floor_heating":
            t_out = 30 - 0.5 * T
        else:
            t_out = self.input_parameters.performance_data["T_out"]

        # Determine delta T
        delta_T = t_out - T

        # Determine COP
        if "AirSourced" in self.name:
            cop = 6.08 - 0.09 * delta_T + 0.0005 * delta_T**2
        elif "GroundSourced" in self.name:
            cop = 10.29 - 0.21 * delta_T + 0.0012 * delta_T**2
        elif "WaterSourced" in self.name:
            cop = 9.97 - 0.20 * delta_T + 0.0012 * delta_T**2

        log.info("Deriving performance data for Heat Pump...")

        if (
            self.component_options.performance_function_type == 1
            or self.component_options.performance_function_type == 2
        ):  # Linear performance function
            size_alpha = 1
        elif self.component_options.performance_function_type == 3:
            size_alpha = 2
        else:
            raise Exception(
                "performance_function_type must be an integer between 1 and 3"
            )

        fit = {}
        fit["out"] = {}
        alpha1 = np.empty(shape=(time_steps, size_alpha))
        alpha2 = np.empty(shape=(time_steps, size_alpha))
        bp_x = np.empty(shape=(time_steps, size_alpha + 1))

        for idx, cop_t in enumerate(cop):
            if idx % 100 == 1:
                print("\rComplete: ", round(idx / time_steps, 2) * 100, "%", end="")

            if self.component_options.performance_function_type == 1:
                x = np.linspace(
                    self.input_parameters.performance_data["min_part_load"], 1, 9
                )
                y = (x / (1 - 0.9 * (1 - x))) * cop_t * x
                coeff = fit_linear_function(x, y)
                alpha1[idx, :] = coeff[0]

            elif self.component_options.performance_function_type == 2:
                x = np.linspace(
                    self.input_parameters.performance_data["min_part_load"], 1, 9
                )
                y = (x / (1 - 0.9 * (1 - x))) * cop_t * x
                x = sm.add_constant(x)
                coeff = fit_linear_function(x, y)
                alpha1[idx, :] = coeff[1]
                alpha2[idx, :] = coeff[0]

            elif (
                self.component_options.performance_function_type == 3
            ):  # piecewise performance function
                y = {}
                x = np.linspace(
                    self.input_parameters.performance_data["min_part_load"], 1, 9
                )
                y["out"] = (x / (1 - 0.9 * (1 - x))) * cop_t * x
                time_step_fit = fit_piecewise_function(x, y, 2)
                alpha1[idx, :] = time_step_fit["out"]["alpha1"]
                alpha2[idx, :] = time_step_fit["out"]["alpha2"]
                bp_x[idx, :] = time_step_fit["out"]["bp_x"]
        print("Complete: ", 100, "%")

        # Coefficients
        fit["coeff"] = {}
        if self.component_options.performance_function_type == 1:
            fit["coeff"]["alpha1"] = alpha1.round(5)

        elif (
            self.component_options.performance_function_type == 2
        ):  # Linear performance function
            fit["coeff"]["alpha1"] = alpha1.round(5)
            fit["coeff"]["alpha2"] = alpha2.round(5)

        elif (
            self.component_options.performance_function_type == 3
        ):  # Piecewise performance function
            fit["coeff"]["alpha1"] = alpha1.round(5)
            fit["coeff"]["alpha2"] = alpha2.round(5)
            fit["coeff"]["bp_x"] = bp_x.round(5)

        # Coefficients
        self.processed_coeff.time_dependent_full = fit["coeff"]

    def _calculate_bounds(self):
        """
        Calculates the bounds of the variables used
        """
        super(HeatPump, self)._calculate_bounds()

        time_steps = len(self.set_t_performance)

        if self.component_options.performance_function_type == 1:
            for c in self.component_options.output_carrier:
                self.bounds["output"][c] = np.column_stack(
                    (
                        np.zeros(shape=(time_steps)),
                        np.ones(shape=(time_steps))
                        * self.processed_coeff.time_dependent_used["alpha1"][:, 0],
                    )
                )

        elif (
            self.component_options.performance_function_type == 2
        ):  # Linear performance function
            for c in self.component_options.output_carrier:
                self.bounds["output"][c] = np.column_stack(
                    (
                        np.zeros(shape=(time_steps)),
                        self.processed_coeff.time_dependent_used["alpha1"][:, -1]
                        + self.processed_coeff.time_dependent_used["alpha2"][:, -1],
                    )
                )

        elif (
            self.component_options.performance_function_type == 3
        ):  # Piecewise performance function
            for c in self.component_options.output_carrier:
                self.bounds["output"][c] = np.column_stack(
                    (
                        np.zeros(shape=(time_steps)),
                        self.processed_coeff.time_dependent_used["alpha1"][:, -1]
                        + self.processed_coeff.time_dependent_used["alpha2"][:, -1],
                    )
                )

        # Input Bounds
        for car in self.component_options.input_carrier:
            self.bounds["input"][car] = np.column_stack(
                (np.zeros(shape=(time_steps)), np.ones(shape=(time_steps)))
            )

        time_steps = len(self.set_t_performance)

    def construct_tech_model(self, b_tec, data: dict, set_t_full, set_t_clustered):
        """
        Adds constraints to technology blocks for tec_type HP (Heat Pump)

        :param b_tec: pyomo block with technology model
        :param dict data: data containing model configuration
        :param set_t_full: pyomo set containing timesteps
        :param set_t_clustered: pyomo set containing clustered timesteps
        :return: pyomo block with technology model
        """
        super(HeatPump, self).construct_tech_model(
            b_tec, data, set_t_full, set_t_clustered
        )

        # DATA OF TECHNOLOGY
        dynamics = self.processed_coeff.dynamics
        rated_power = self.input_parameters.rated_power

        if self.component_options.performance_function_type == 1:
            b_tec = self._performance_function_type_1(b_tec)
        elif self.component_options.performance_function_type == 2:
            b_tec = self._performance_function_type_2(b_tec)
            self.big_m_transformation_required = 1
        elif self.component_options.performance_function_type == 3:
            b_tec = self._performance_function_type_3(b_tec)
            self.big_m_transformation_required = 1

        # size constraint based on input
        def init_size_constraint(const, t):
            return self.input[t, "electricity"] <= b_tec.var_size * rated_power

        b_tec.const_size = pyo.Constraint(
            self.set_t_performance, rule=init_size_constraint
        )

        # RAMPING RATES
        if "ramping_time" in dynamics:
            if not dynamics["ramping_time"] == -1:
                b_tec = self._define_ramping_rates(b_tec, data)

        return b_tec

    def _performance_function_type_1(self, b_tec):
        """
        Linear, no minimal partload, through origin
        :param b_tec: technology block
        :return: technology block
        """
        # Get performance parameters
        coeff_td = self.processed_coeff.time_dependent_used
        alpha1 = coeff_td["alpha1"]

        def init_input_output(const, t):
            return (
                self.output[t, "heat"] == alpha1[t - 1] * self.input[t, "electricity"]
            )

        b_tec.const_input_output = pyo.Constraint(
            self.set_t_performance, rule=init_input_output
        )

        return b_tec

    def _performance_function_type_2(self, b_tec):
        """
        Linear, minimal partload
        :param b_tec: technology block
        :return: technology block
        """
        # Get performance parameters
        coeff_td = self.processed_coeff.time_dependent_used
        coeff_ti = self.processed_coeff.time_independent
        alpha1 = coeff_td["alpha1"]
        alpha2 = coeff_td["alpha2"]
        min_part_load = coeff_ti["min_part_load"]
        rated_power = self.input_parameters.rated_power

        # define disjuncts for on/off
        s_indicators = range(0, 2)

        def init_input_output(dis, t, ind):
            if ind == 0:  # technology off

                def init_input_off(const):
                    return self.input[t, "electricity"] == 0

                dis.const_input = pyo.Constraint(rule=init_input_off)

                def init_output_off(const):
                    return self.output[t, "heat"] == 0

                dis.const_output_off = pyo.Constraint(rule=init_output_off)
            else:  # technology on
                # input-output relation
                def init_input_output_on(const):
                    return (
                        self.output[t, "heat"]
                        == alpha1[t - 1] * self.input[t, "electricity"]
                        + alpha2[t - 1] * b_tec.var_size * rated_power
                    )

                dis.const_input_output_on = pyo.Constraint(rule=init_input_output_on)

                # min part load relation
                def init_min_partload(const):
                    return (
                        self.input[t, "electricity"]
                        >= min_part_load * b_tec.var_size * rated_power
                    )

                dis.const_min_partload = pyo.Constraint(rule=init_min_partload)

        b_tec.dis_input_output = gdp.Disjunct(
            self.set_t_performance, s_indicators, rule=init_input_output
        )

        # Bind disjuncts
        def bind_disjunctions(dis, t):
            return [b_tec.dis_input_output[t, i] for i in s_indicators]

        b_tec.disjunction_input_output = gdp.Disjunction(
            self.set_t_performance, rule=bind_disjunctions
        )

        return b_tec

    def _performance_function_type_3(self, b_tec):
        """
        Piece-wise linear, minimal partload
        :param b_tec: technology block
        :return: technology block
        """
        # Get performance parameters
        coeff_td = self.processed_coeff.time_dependent_used
        coeff_ti = self.processed_coeff.time_independent
        alpha1 = coeff_td["alpha1"]
        alpha2 = coeff_td["alpha2"]
        bp_x = coeff_td["bp_x"]
        min_part_load = coeff_ti["min_part_load"]
        rated_power = self.input_parameters.rated_power

        s_indicators = range(0, 2)

        def init_input_output(dis, t, ind):
            if ind == 0:  # technology off

                def init_input_off(const):
                    return self.input[t, "electricity"] == 0

                dis.const_input_off = pyo.Constraint(rule=init_input_off)

                def init_output_off(const):
                    return self.output[t, "heat"] == 0

                dis.const_output_off = pyo.Constraint(rule=init_output_off)

            else:  # piecewise definition

                def init_input_on1(const):
                    return (
                        self.input[t, "electricity"]
                        >= bp_x[t - 1, ind] * b_tec.var_size * rated_power
                    )

                dis.const_input_on1 = pyo.Constraint(rule=init_input_on1)

                def init_input_on2(const):
                    return (
                        self.input[t, "electricity"]
                        <= bp_x[t - 1, ind + 1] * b_tec.var_size * rated_power
                    )

                dis.const_input_on2 = pyo.Constraint(rule=init_input_on2)

                def init_output_on(const):
                    return (
                        self.output[t, "heat"]
                        == alpha1[t - 1, ind - 1] * self.input[t, "electricity"]
                        + alpha2[t - 1, ind - 1] * b_tec.var_size * rated_power
                    )

                dis.const_input_output_on = pyo.Constraint(rule=init_output_on)

                # min part load relation
                def init_min_partload(const):
                    return (
                        self.input[t, "electricity"]
                        >= min_part_load * b_tec.var_size * rated_power
                    )

                dis.const_min_partload = pyo.Constraint(rule=init_min_partload)

        b_tec.dis_input_output = gdp.Disjunct(
            self.set_t_performance, s_indicators, rule=init_input_output
        )

        # Bind disjuncts
        def bind_disjunctions(dis, t):
            return [b_tec.dis_input_output[t, i] for i in s_indicators]

        b_tec.disjunction_input_output = gdp.Disjunction(
            self.set_t_performance, rule=bind_disjunctions
        )

        return b_tec

    def _define_ramping_rates(self, b_tec, data):
        """
        Constraints the inputs for a ramping rate

        :param b_tec: technology model block
        :return:
        """
        dynamics = self.processed_coeff.dynamics

        ramping_time = dynamics["ramping_time"]

        # Calculate ramping rates
        if "ref_size" in dynamics and not dynamics["ref_size"] == -1:
            ramping_rate = dynamics["ref_size"] / ramping_time
        else:
            ramping_rate = b_tec.var_size / ramping_time

        # Constraints ramping rates
        if "ramping_const_int" in dynamics and dynamics["ramping_const_int"] == 1:

            s_indicators = range(0, 3)

            def init_ramping_operation_on(dis, t, ind):
                if t > 1:
                    if ind == 0:  # ramping constrained
                        dis.const_ramping_on = pyo.Constraint(
                            expr=b_tec.var_x[t] - b_tec.var_x[t - 1] == 0
                        )

                        def init_ramping_down_rate_operation(const):
                            return -ramping_rate <= sum(
                                self.input[t, car_input] - self.input[t - 1, car_input]
                                for car_input in b_tec.set_input_carriers
                            )

                        dis.const_ramping_down_rate = pyo.Constraint(
                            rule=init_ramping_down_rate_operation
                        )

                        def init_ramping_up_rate_operation(const):
                            return (
                                sum(
                                    self.input[t, car_input]
                                    - self.input[t - 1, car_input]
                                    for car_input in b_tec.set_input_carriers
                                )
                                <= ramping_rate
                            )

                        dis.const_ramping_up_rate = pyo.Constraint(
                            rule=init_ramping_up_rate_operation
                        )

                    elif ind == 1:  # startup, no ramping constraint
                        dis.const_ramping_on = pyo.Constraint(
                            expr=b_tec.var_x[t] - b_tec.var_x[t - 1] == 1
                        )

                    else:  # shutdown, no ramping constraint
                        dis.const_ramping_on = pyo.Constraint(
                            expr=b_tec.var_x[t] - b_tec.var_x[t - 1] == -1
                        )

            b_tec.dis_ramping_operation_on = gdp.Disjunct(
                self.set_t_performance, s_indicators, rule=init_ramping_operation_on
            )

            # Bind disjuncts
            def bind_disjunctions(dis, t):
                return [b_tec.dis_ramping_operation_on[t, i] for i in s_indicators]

            b_tec.disjunction_ramping_operation_on = gdp.Disjunction(
                self.set_t_performance, rule=bind_disjunctions
            )

        else:

            if data["config"]["optimization"]["typicaldays"]["N"]["value"] == 0:
                input_aux_rr = self.input
                set_t_rr = self.set_t_performance
            else:
                if (
                    data["config"]["optimization"]["typicaldays"]["method"]["value"]
                    == 1
                ):
                    sequence = data["k_means_specs"]["sequence"]
                elif (
                    data["config"]["optimization"]["typicaldays"]["method"]["value"]
                    == 2
                ):
                    sequence = self.sequence

                # init bounds at full res
                bounds_rr_full = {
                    "input": self.fitting_class.calculate_input_bounds(
                        self.component_options.size_based_on, len(self.set_t_full)
                    )
                }

                # create input variable for full res
                def init_input_bounds(bounds, t, car):
                    return tuple(
                        bounds_rr_full["input"][car][t - 1, :]
                        * self.processed_coeff.time_independent["size_max"]
                        * self.processed_coeff.time_independent["rated_power"]
                    )

                b_tec.var_input_rr_full = pyo.Var(
                    self.set_t_full,
                    b_tec.set_input_carriers,
                    within=pyo.NonNegativeReals,
                    bounds=init_input_bounds,
                )

                b_tec.const_link_full_resolution_rr = link_full_resolution_to_clustered(
                    self.input,
                    b_tec.var_input_rr_full,
                    self.set_t_full,
                    sequence,
                    b_tec.set_input_carriers,
                )

                input_aux_rr = b_tec.var_input_rr_full
                set_t_rr = self.set_t_full

            # Ramping constraint without integers
            def init_ramping_down_rate(const, t):
                if t > 1:
                    return -ramping_rate <= sum(
                        input_aux_rr[t, car_input] - input_aux_rr[t - 1, car_input]
                        for car_input in b_tec.set_input_carriers
                    )
                else:
                    return pyo.Constraint.Skip

            b_tec.const_ramping_down_rate = pyo.Constraint(
                set_t_rr, rule=init_ramping_down_rate
            )

            def init_ramping_up_rate(const, t):
                if t > 1:
                    return (
                        sum(
                            input_aux_rr[t, car_input] - input_aux_rr[t - 1, car_input]
                            for car_input in b_tec.set_input_carriers
                        )
                        <= ramping_rate
                    )
                else:
                    return pyo.Constraint.Skip

            b_tec.const_ramping_up_rate = pyo.Constraint(
                set_t_rr, rule=init_ramping_up_rate
            )

        return b_tec



import pandas as pd
import pyomo.environ as pyo
import h5py
import numpy as np

from ..technology import Technology
from ....components.utilities import (
    annualize,
    set_discount_rate,
    link_full_resolution_to_clustered,
)
from ...component import InputParameters


class Sink(Technology):
    """
    Permanent storage technology (has no output)

    This model resembles a permanent storage technology (sink). It takes energy and a main carrier (e.g. CO2 etc)
    as inputs, and it has no output.

    **Variable declarations:**

    - ``var_storage_level``: Storage level in :math:`t`: :math:`E_t`

    - ``var_injection_capacity``: Injection capacity :math: `injCapacity`

    **Constraint declarations:**

    - Size constraint:

      .. math::
        E_{t} \\leq S

    - Maximal injection rate:

      .. math::
        Input_{t, maincar} \\leq injCapacity

    - Maximal injection capacity:

      .. math::
        injCapacity \\leq injRateMax


    - Storage level calculation:

      .. math::
        E_{t} = E_{t-1} + Input_{t}

    - If an energy consumption for the injection is given, the respective carrier input is:

      .. math::
        Input_{t, car} = cons_{car, in} Input_{t}

    - CAPEX is given by two contributions

        .. math::
            CAPEX_{storSize} = Size_{storSize} * UnitCost_{storSize}
            CAPEX_{injCapacity} = injCapacity * UnitCost_{injCapacity}

    - Additionally, ramping rates of the technology can be constraint (for input and
      output).

      .. math::
         -rampingrate \\leq Input_{t, maincar} - Input_{t-1, maincar} \\leq rampingrate

    """

    def __init__(self, tec_data: dict):
        """
        Constructor

        :param dict tec_data: technology data
        """
        super().__init__(tec_data)

        self.component_options.emissions_based_on = "input"
        self.component_options.main_input_carrier = tec_data["Performance"][
            "main_input_carrier"
        ]
        self.flexibility_data = tec_data["Flexibility"]

    def fit_technology_performance(self, climate_data: pd.DataFrame, location: dict):
        """
        Calculate input bounds and select new capex model

        :param pd.Dataframe climate_data: dataframe containing climate data
        :param dict location: dict containing location details
        """
        super(Sink, self).fit_technology_performance(climate_data, location)

        # For a flexibly optimized storage technology (i.e., not a fixed P-E ratio), an adapted CAPEX function is used
        # to account for charging and discharging capacity costs.
        if self.flexibility_data["injection_capacity_is_decision_var"]:
            self.economics.capex_model = 4

        self.processed_coeff.time_independent["injection_rate_max"] = (
            self.flexibility_data["injection_rate_max"]
        )
        if (
            "energy_consumption"
            in self.input_parameters.performance_data["performance"]
        ):
            self.processed_coeff.time_independent["energy_consumption"] = (
                self.input_parameters.performance_data["performance"][
                    "energy_consumption"
                ]
            )

    def _calculate_bounds(self):
        """
        Calculates the bounds of the variables used
        """
        super(Sink, self)._calculate_bounds()

        time_steps = len(self.set_t_performance)

        # Input Bounds
        for car in self.component_options.input_carrier:
            if car == self.component_options.main_input_carrier:
                self.bounds["input"][car] = np.column_stack(
                    (
                        np.zeros(shape=(time_steps)),
                        np.ones(shape=(time_steps))
                        * self.flexibility_data["injection_rate_max"],
                    )
                )
            else:
                if (
                    "energy_consumption"
                    in self.input_parameters.performance_data["performance"]
                ):
                    energy_consumption = self.input_parameters.performance_data[
                        "performance"
                    ]["energy_consumption"]
                    self.bounds["input"][car] = np.column_stack(
                        (
                            np.zeros(shape=(time_steps)),
                            np.ones(shape=(time_steps))
                            * self.flexibility_data["injection_rate_max"]
                            * energy_consumption["in"][car],
                        )
                    )

    def construct_tech_model(self, b_tec, data: dict, set_t_full, set_t_clustered):
        """
        Construct SINK constraints

        Adds constraints to technology blocks for tec_type SINK, resembling a permanent storage technology

        :param b_tec: pyomo block with technology model
        :param dict data: data containing model configuration
        :param set_t_full: pyomo set containing timesteps
        :param set_t_clustered: pyomo set containing clustered timesteps
        :return: pyomo block with technology model
        """

        super(Sink, self).construct_tech_model(b_tec, data, set_t_full, set_t_clustered)

        # DATA OF TECHNOLOGY
        config = data["config"]
        coeff_ti = self.processed_coeff.time_independent
        dynamics = self.processed_coeff.dynamics

        # sequence_storage = self.sequence
        if config["optimization"]["typicaldays"]["N"]["value"] == 0:
            sequence_storage = self.sequence
        elif config["optimization"]["typicaldays"]["method"]["value"] == 1:
            sequence_storage = data["k_means_specs"]["sequence"]
        elif config["optimization"]["typicaldays"]["method"]["value"] == 2:
            sequence_storage = self.sequence

        # Sotrage level and injection capacity decision variables
        b_tec.var_storage_level = pyo.Var(
            set_t_full,
            domain=pyo.NonNegativeReals,
            bounds=(b_tec.para_size_min, b_tec.para_size_max),
        )
        b_tec.var_injection_capacity = pyo.Var(
            domain=pyo.NonNegativeReals,
            bounds=(0, coeff_ti["injection_rate_max"]),
        )

        if self.flexibility_data["injection_capacity_is_decision_var"]:
            b_tec = self._define_sink_capex(b_tec, data)

        # Maximum storage level constraint
        def init_size_constraint(const, t):
            # storageLevel <= storSize
            return b_tec.var_storage_level[t] <= b_tec.var_size

        b_tec.const_size = pyo.Constraint(set_t_full, rule=init_size_constraint)

        # Constraint storage level

        def init_storage_level(const, t):
            # storageLevel[1] <= injRate[1]
            if t == 1:
                return (
                    b_tec.var_storage_level[t]
                    == self.input[
                        sequence_storage[t - 1],
                        self.component_options.main_input_carrier,
                    ]
                )
            else:
                # storageLevel[t] <= storageLevel[t-1]+injRate[t]
                return (
                    b_tec.var_storage_level[t]
                    == b_tec.var_storage_level[t - 1]
                    + self.input[
                        sequence_storage[t - 1],
                        self.component_options.main_input_carrier,
                    ]
                )

        b_tec.const_storage_level = pyo.Constraint(set_t_full, rule=init_storage_level)

        # Maximal injection rate
        def init_maximal_injection(const, t):
            # input[t] <= injectionCapacity
            return (
                self.input[t, self.component_options.main_input_carrier]
                <= b_tec.var_injection_capacity
            )

        b_tec.const_max_injection = pyo.Constraint(
            self.set_t_performance, rule=init_maximal_injection
        )

        # if injection rates are fixed/ flexible:
        def init_max_capacity_injection(const):
            if self.flexibility_data["injection_capacity_is_decision_var"]:
                # injectionCapacity <= injectionRateMax
                return b_tec.var_injection_capacity <= coeff_ti["injection_rate_max"]
            else:
                # injectionCapacity == injectionRateMax
                return b_tec.var_injection_capacity == coeff_ti["injection_rate_max"]

        b_tec.const_max_injection_cap = pyo.Constraint(rule=init_max_capacity_injection)

        # Energy consumption for injection
        if "energy_consumption" in coeff_ti:
            energy_consumption = coeff_ti["energy_consumption"]
            if "in" in energy_consumption:
                b_tec.set_energyconsumption_carriers_in = pyo.Set(
                    initialize=list(energy_consumption["in"].keys())
                )

                def init_energyconsumption_in(const, t, car):
                    # energyInput[t] = mainInput[t] * energyConsumption
                    return (
                        self.input[t, car]
                        == self.input[t, self.component_options.main_input_carrier]
                        * energy_consumption["in"][car]
                    )

                b_tec.const_energyconsumption_in = pyo.Constraint(
                    self.set_t_performance,
                    b_tec.set_energyconsumption_carriers_in,
                    rule=init_energyconsumption_in,
                )

        # RAMPING RATES
        if "ramping_time" in dynamics:
            if not dynamics["ramping_time"] == -1:
                b_tec = self._define_ramping_rates(b_tec, data)

        return b_tec

    def _define_sink_capex(self, b_tec, data: dict):
        """

        Construct CAPEX of SINK constraints

        Adds constraints to technology blocks for tec_type SINK to calculate the CAPEX

        :param b_tec: pyomo block with technology model
        :param dict data: data containing model configuration
        :return: pyomo block with technology model
        """

        config = data["config"]
        economics = self.economics
        discount_rate = set_discount_rate(config, economics)
        fraction_of_year_modelled = data["topology"]["fraction_of_year_modelled"]
        annualization_factor = annualize(
            discount_rate, economics.lifetime, fraction_of_year_modelled
        )
        flexibility = self.flexibility_data
        coeff_ti = self.processed_coeff.time_independent

        # CAPEX PARAMETERS
        b_tec.para_unit_capex_injection_cap = pyo.Param(
            domain=pyo.Reals,
            initialize=flexibility["capex_injection_cap"],
            mutable=True,
        )
        b_tec.para_unit_capex_stor_size = pyo.Param(
            domain=pyo.Reals,
            initialize=economics.capex_data["unit_capex"],
            mutable=True,
        )

        b_tec.para_unit_capex_injection_cap_annual = pyo.Param(
            domain=pyo.Reals,
            initialize=(annualization_factor * b_tec.para_unit_capex_injection_cap),
            mutable=True,
        )
        b_tec.para_unit_capex_stor_size_annual = pyo.Param(
            domain=pyo.Reals,
            initialize=(annualization_factor * b_tec.para_unit_capex_stor_size),
            mutable=True,
        )

        # BOUNDS
        max_capex_injection_cap = (
            b_tec.para_unit_capex_injection_cap_annual * coeff_ti["injection_rate_max"]
        )
        max_capex_stor_size = (
            b_tec.para_unit_capex_stor_size_annual * b_tec.para_size_max
        )

        b_tec.var_capex_injection_cap = pyo.Var(
            domain=pyo.NonNegativeReals, bounds=(0, max_capex_injection_cap)
        )
        b_tec.var_capex_stor_size = pyo.Var(
            domain=pyo.NonNegativeReals, bounds=(0, max_capex_stor_size)
        )

        # CAPEX constraints
        # CAPEXinjection = injCapacity * UnitCost_injCapacity
        b_tec.const_capex_injection_cap = pyo.Constraint(
            expr=b_tec.var_injection_capacity
            * b_tec.para_unit_capex_injection_cap_annual
            == b_tec.var_capex_injection_cap
        )
        # CAPEXstorSize = storSize * UnitCost_storSize
        b_tec.const_capex_stor_size = pyo.Constraint(
            expr=b_tec.var_size * b_tec.para_unit_capex_stor_size_annual
            == b_tec.var_capex_stor_size
        )
        b_tec.const_capex_aux = pyo.Constraint(
            expr=b_tec.var_capex_injection_cap + b_tec.var_capex_stor_size
            == b_tec.var_capex_aux
        )

        return b_tec

    def write_results_tec_operation(self, h5_group: h5py.Group, model_block: pyo.Block):
        """
        Function to report results of technologies operations after optimization

        :param Block b_tec: technology model block
        :param h5py.Group h5_group: technology model block
        """
        super(Sink, self).write_results_tec_operation(h5_group, model_block)

        h5_group.create_dataset(
            "storage_level",
            data=[model_block.var_storage_level[t].value for t in self.set_t_full],
        )

    def write_results_tec_design(self, h5_group: h5py.Group, model_block: pyo.Block):
        """
        Function to report results of technologies design after optimization

        :param  h5py.Group h5_group: h5 file structure
        :param Block b_tec: technology model block
        """
        super(Sink, self).write_results_tec_design(h5_group, model_block)

        if self.flexibility_data["injection_capacity_is_decision_var"]:
            h5_group.create_dataset(
                "injection_capacity", data=[model_block.var_injection_capacity.value]
            )
            h5_group.create_dataset(
                "capex_injection_cap", data=[model_block.var_capex_injection_cap.value]
            )
            h5_group.create_dataset(
                "capex_stor_size", data=[model_block.var_capex_stor_size.value]
            )

    def _define_ramping_rates(self, b_tec, data):
        """
        Constraints the inputs for a ramping rate

        :param Block b_tec: technology model block
        :return: technology Block with the constraints for the dynamic behaviour
        """
        dynamics = self.processed_coeff.dynamics

        ramping_time = dynamics["ramping_time"]

        # Calculate ramping rates
        if "ref_size" in dynamics and not dynamics["ref_size"] == -1:
            ramping_rate = dynamics["ref_size"] / ramping_time
        else:
            ramping_rate = b_tec.var_size / ramping_time

        if data["config"]["optimization"]["typicaldays"]["N"]["value"] == 0:
            input_aux_rr = self.input
            set_t_rr = self.set_t_performance
        else:
            if data["config"]["optimization"]["typicaldays"]["method"]["value"] == 1:
                sequence = data["k_means_specs"]["sequence"]
            elif data["config"]["optimization"]["typicaldays"]["method"]["value"] == 2:
                sequence = self.sequence

            # init bounds at full res
            bounds_rr_full = {
                "input": self.fitting_class.calculate_input_bounds(
                    self.component_options.size_based_on, len(self.set_t_full)
                )
            }

            # create input variable for full res
            def init_input_bounds(bounds, t, car):
                return tuple(
                    bounds_rr_full["input"][car][t - 1, :]
                    * self.processed_coeff.time_independent["size_max"]
                    * self.processed_coeff.time_independent["rated_power"]
                )

            b_tec.var_input_rr_full = pyo.Var(
                self.set_t_full,
                b_tec.set_input_carriers,
                within=pyo.NonNegativeReals,
                bounds=init_input_bounds,
            )

            b_tec.const_link_full_resolution_rr = link_full_resolution_to_clustered(
                self.input,
                b_tec.var_input_rr_full,
                self.set_t_full,
                sequence,
                b_tec.set_input_carriers,
            )

            input_aux_rr = b_tec.var_input_rr_full
            set_t_rr = self.set_t_full

        # Ramping constraint without integers
        def init_ramping_down_rate(const, t):
            if t > 1:
                return (
                    -ramping_rate
                    <= input_aux_rr[t, self.component_options.main_input_carrier]
                    - input_aux_rr[t - 1, self.component_options.main_input_carrier]
                )
            else:
                return pyo.Constraint.Skip

        b_tec.const_ramping_down_rate = pyo.Constraint(
            set_t_rr, rule=init_ramping_down_rate
        )

        def init_ramping_up_rate(const, t):
            if t > 1:
                return (
                    input_aux_rr[t, self.component_options.main_input_carrier]
                    - input_aux_rr[t - 1, self.component_options.main_input_carrier]
                    <= ramping_rate
                )
            else:
                return pyo.Constraint.Skip

        b_tec.const_ramping_up_rate = pyo.Constraint(
            set_t_rr, rule=init_ramping_up_rate
        )

        return b_tec



import h5py
from pathlib import Path
import os

from pyomo.environ import ConcreteModel
from ..utilities import get_set_t

import logging

log = logging.getLogger(__name__)


def get_summary(model, solution, folder_path: Path, model_info: dict) -> dict:
    """
    Retrieves all variable values relevant for the summary of an optimization run.

    These variables and their values are written to a dictionary.

    :param model: the model for which you want to obtain the results summary.
    :param solution: Pyomo solver results
    :param Path folder_path: folder path of optimization run
    :param dict model_info: information of the last solve done by the model
    :return: a dictionary containing the most important model results (i.e., summary_dict)
    :rtype: dict
    """
    # SUMMARY: create dictionary
    summary_dict = {}

    # summary: retrieve / calculate cost variables
    summary_dict["total_npv"] = model.var_npv.value
    summary_dict["cost_capex_tecs"] = sum(
        model.periods[period].var_cost_capex_tecs.value for period in model.set_periods
    )

    summary_dict["cost_capex_netws"] = sum(
        model.periods[period].var_cost_capex_netws.value for period in model.set_periods
    )
    summary_dict["cost_opex_tecs"] = sum(
        model.periods[period].var_cost_opex_tecs.value for period in model.set_periods
    )
    summary_dict["cost_opex_netws"] = sum(
        model.periods[period].var_cost_opex_netws.value for period in model.set_periods
    )
    summary_dict["cost_tecs"] = sum(
        model.periods[period].var_cost_tecs.value for period in model.set_periods
    )
    summary_dict["cost_netws"] = sum(
        model.periods[period].var_cost_netws.value for period in model.set_periods
    )
    summary_dict["cost_imports"] = sum(
        model.periods[period].var_cost_imports.value for period in model.set_periods
    )
    summary_dict["cost_exports"] = sum(
        model.periods[period].var_cost_exports.value for period in model.set_periods
    )
    summary_dict["violation_cost"] = sum(
        model.periods[period].var_cost_violation.value for period in model.set_periods
    )
    summary_dict["carbon_revenue"] = sum(
        model.periods[period].var_carbon_revenue.value for period in model.set_periods
    )
    summary_dict["carbon_cost"] = sum(
        model.periods[period].var_carbon_cost.value for period in model.set_periods
    )
    summary_dict["total_cost"] = sum(
        model.periods[period].var_cost_total.value for period in model.set_periods
    )

    summary_dict["emissions_pos"] = sum(
        model.periods[period].var_emissions_pos.value for period in model.set_periods
    )
    summary_dict["emissions_neg"] = sum(
        model.periods[period].var_emissions_neg.value for period in model.set_periods
    )
    summary_dict["emissions_net"] = sum(
        model.periods[period].var_emissions_net.value for period in model.set_periods
    )

    # summary: retrieve / calculate solver status
    try:
        int(solution.solver(0).wallclock_time)
        time = solution.solver(0).wallclock_time
    except:
        time = 0
    summary_dict["time_total"] = time
    summary_dict["lb"] = solution.problem(0).lower_bound
    summary_dict["ub"] = solution.problem(0).upper_bound
    summary_dict["absolute gap"] = (
        solution.problem(0).upper_bound - solution.problem(0).lower_bound
    )
    summary_dict["solver_status"] = solution.solver.termination_condition.value

    # summary: retrieve / calculate run specs
    summary_dict["objective"] = model_info["config"]["optimization"]["objective"][
        "value"
    ]
    summary_dict["pareto_point"] = model_info["pareto_point"]
    summary_dict["monte_carlo_run"] = model_info["monte_carlo_run"]
    summary_dict["time_stage"] = model_info["time_stage"]

    summary_dict["case"] = model_info["config"]["reporting"]["case_name"]["value"]

    summary_dict["time_stamp"] = str(folder_path)

    return summary_dict


def write_optimization_results_to_h5(model, solution, model_info: dict, data) -> dict:
    """
    Collects the results from the model blocks and writes them to an HDF5 file

    Saving to HDF5 files is done using the h5py library.
    The summary results are returned in a dictionary format for exporting to Excel.
    Overhead (calculation of variables) are placed in the utilities file.

    :param ConcreteModel model: the model for which you want to save the results to an HDF5 file.
    :param solution: Pyomo solver results
    :param dict model_info: information of the last solve done by the model
    :param data: DataHandle object containing all data read in by the DataHandle class.
    :return: a dictionary containing the most important model results (i.e., summary_dict)
    :rtype: dict
    """

    def read_value_from_parameter(para) -> list:
        """
        Reads parameter values from a pyomo parameter

        :param para: pyomo parameter
        :return: values as a list
        :rtype: list
        """
        if para.mutable:
            return [para[t, car].value for t in set_t]
        else:
            return [para[t, car] for t in set_t]

    config = model_info["config"]
    folder_path = model_info["result_folder_path"]

    # LOG
    log_msg = f"Writing results to {folder_path}"
    log.info(log_msg)

    # create the results h5 file in the results folder
    h5_file_path = os.path.join(folder_path, "optimization_results.h5")
    with h5py.File(h5_file_path, mode="w") as f:

        summary_dict = get_summary(model, solution, folder_path, model_info)

        # SUMMARY [g]: convert dictionary to h5 datasets
        summary = f.create_group("summary")
        for key in summary_dict:
            if summary_dict[key] is None:
                value = -1
            else:
                value = summary_dict[key]
            summary.create_dataset(key, data=value)

        # TIME AGGREGATION INFORMATION [g]:
        # K-means specs
        k_means_specs = f.create_group("k_means_specs")
        for investment_period in data.k_means_specs:
            k_means_specs_period = k_means_specs.create_group(investment_period)
            for key in data.k_means_specs[investment_period]:
                k_means_specs_period.create_dataset(
                    key, data=data.k_means_specs[investment_period][key]
                )

        # Topology Information
        topology = f.create_group("topology")
        topology.create_dataset("nodes", data=list(model.set_nodes))
        topology.create_dataset("periods", data=list(model.set_periods))
        topology.create_dataset("carriers", data=list(model.set_carriers))

        # TIME-INDEPENDENT RESULTS (design) [g]
        g_design = f.create_group("design")

        # TIME-INDEPENDENT RESULTS: NETWORKS [g] > within: specific network [g] > within: specific arc of network[g]
        networks_design = g_design.create_group("networks")

        for period in model.set_periods:
            g_period_netw_design = networks_design.create_group(period)

            b_period = model.periods[period]
            set_t = get_set_t(config, b_period)

            if not config["energybalance"]["copperplate"]["value"]:
                for netw_name in b_period.set_networks:
                    netw_specific_group = g_period_netw_design.create_group(netw_name)
                    b_netw = b_period.network_block[netw_name]
                    data.network_data[period][netw_name].write_results_netw_design(
                        netw_specific_group, b_netw
                    )

        # TIME-INDEPENDENT RESULTS: NODES [g]
        nodes_design = g_design.create_group("nodes")
        for period in model.set_periods:
            g_period_node_design = nodes_design.create_group(period)

            # TIME-INDEPENDENT RESULTS: NODES: specific node [g] within: specific technology [g]
            for node_name in model.set_nodes:
                node_specific_group = g_period_node_design.create_group(node_name)
                b_node = b_period.node_blocks[node_name]

                for tec_name in b_node.set_technologies:
                    tec_group = node_specific_group.create_group(tec_name)
                    b_tec = b_node.tech_blocks_active[tec_name]
                    data.technology_data[period][node_name][
                        tec_name
                    ].write_results_tec_design(tec_group, b_tec)

        # TIME-DEPENDENT RESULTS (operation) [g]
        operation = f.create_group("operation")

        # TIME-DEPENDENT RESULTS: NETWORKS [g] > within: specific network [g] > within: specific arc of network [g]
        networks_operation = operation.create_group("networks")

        for period in model.set_periods:
            g_period_netw_operation = networks_operation.create_group(period)

            if not config["energybalance"]["copperplate"]["value"]:
                for netw_name in b_period.set_networks:
                    netw_specific_group = g_period_netw_operation.create_group(
                        netw_name
                    )
                    b_netw = b_period.network_block[netw_name]
                    data.network_data[period][netw_name].write_results_netw_operation(
                        netw_specific_group, b_netw
                    )

        # TECHNOLOGY OPERATION [g] > within: node > specific technology [g]
        tec_operation_group = operation.create_group("technology_operation")
        for period in model.set_periods:
            g_period_tec_operation = tec_operation_group.create_group(period)

            for node_name in model.set_nodes:
                node_specific_group = g_period_tec_operation.create_group(node_name)
                b_node = b_period.node_blocks[node_name]

                for tec_name in b_node.set_technologies:
                    tec_group = node_specific_group.create_group(tec_name)
                    b_tec = b_node.tech_blocks_active[tec_name]
                    data.technology_data[period][node_name][
                        tec_name
                    ].write_results_tec_operation(tec_group, b_tec)

        # ENERGY BALANCE [g] > within: node > specific carrier [g]
        ebalance_group = operation.create_group("energy_balance")

        for period in model.set_periods:
            g_period_ebalance = ebalance_group.create_group(period)

            for node_name in model.set_nodes:
                node_specific_group = g_period_ebalance.create_group(node_name)
                b_node = b_period.node_blocks[node_name]

                for car in b_node.set_carriers:
                    car_group = node_specific_group.create_group(car)
                    node_data = b_node

                    technology_inputs = [
                        sum(
                            node_data.tech_blocks_active[tec].var_input[t, car].value
                            for tec in node_data.set_technologies
                            if car
                            in node_data.tech_blocks_active[tec].set_input_carriers
                        )
                        for t in set_t
                    ]
                    car_group.create_dataset(
                        "technology_inputs", data=technology_inputs
                    )
                    technology_outputs = [
                        sum(
                            node_data.tech_blocks_active[tec].var_output[t, car].value
                            for tec in node_data.set_technologies
                            if car
                            in node_data.tech_blocks_active[tec].set_output_carriers
                        )
                        for t in set_t
                    ]
                    car_group.create_dataset(
                        "technology_outputs", data=technology_outputs
                    )
                    car_group.create_dataset(
                        "generic_production",
                        data=[
                            node_data.var_generic_production[t, car].value
                            for t in set_t
                        ],
                    )
                    car_group.create_dataset(
                        "network_inflow",
                        data=[
                            0 if x is None else x
                            for x in [
                                node_data.var_netw_inflow[t, car].value for t in set_t
                            ]
                        ],
                    )
                    car_group.create_dataset(
                        "network_outflow",
                        data=[
                            0 if x is None else x
                            for x in [
                                node_data.var_netw_outflow[t, car].value for t in set_t
                            ]
                        ],
                    )
                    if hasattr(node_data, "var_netw_consumption"):
                        network_consumption = [
                            node_data.var_netw_consumption[t, car].value for t in set_t
                        ]
                        car_group.create_dataset(
                            "network_consumption", data=network_consumption
                        )
                    car_group.create_dataset(
                        "import",
                        data=[node_data.var_import_flow[t, car].value for t in set_t],
                    )

                    car_group.create_dataset(
                        "import_price",
                        data=read_value_from_parameter(node_data.para_import_price),
                    )
                    car_group.create_dataset(
                        "export",
                        data=[node_data.var_export_flow[t, car].value for t in set_t],
                    )
                    car_group.create_dataset(
                        "export_price",
                        data=read_value_from_parameter(node_data.para_export_price),
                    )
                    car_group.create_dataset(
                        "demand", data=[node_data.para_demand[t, car] for t in set_t]
                    )

    return summary_dict



import pyomo.environ as pyo
import pyomo.gdp as gdp
import numpy as np
import pandas as pd

from ...component import InputParameters
from ..technology import Technology


class Conv4(Technology):
    """
    Technology with no inputs

    This technology type resembles a technology with fixed output ratios  and no
    inputs, i.e., :math:`output_{car} \\leq S`. This technology is useful for
    modelling a technology for which you do not care about the inputs, i.e., you do not
    wish to construct and solve an energy balance for the input carriers.
    Two different performance function fits are possible.

    **Constraint declarations:**

    For all performance function types, the following constraints hold:

    - Size constraints are formulated on the output.

      .. math::
         Output_{t, maincarrier} \\leq S

    - The ratios of outputs are fixed and given as:

      .. math::
        Output_{t, car} = {\\phi}_{car} * Output_{t, maincarrier}

    - ``performance_function_type == 1``: No further constraints on the performance
      of the technology.

    - ``performance_function_type == 2``: A minimum part load can be specified (
      requiring a big-m transformation for the solving). The following constraints hold:

      When the technology is on:

      .. math::
        Output_{maincarrier} \\geq Output_{min} * S

      When the technology is off, output is set to 0:

      .. math::
         Output_{t, car} = 0
    """

    def __init__(self, tec_data: dict):
        """
        Constructor

        :param dict tec_data: technology data
        """
        super().__init__(tec_data)

        self.component_options.emissions_based_on = "output"
        self.component_options.main_output_carrier = tec_data["Performance"][
            "main_output_carrier"
        ]

    def fit_technology_performance(self, climate_data: pd.DataFrame, location: dict):
        """
        Fits conversion technology type 4

        :param pd.Dataframe climate_data: dataframe containing climate data
        :param dict location: dict containing location details
        """
        super(Conv4, self).fit_technology_performance(climate_data, location)

        # Coefficients
        phi = {}
        for car in self.input_parameters.performance_data["output_ratios"]:
            phi[car] = self.input_parameters.performance_data["output_ratios"][car]
        self.processed_coeff.time_independent["phi"] = phi

    def _calculate_bounds(self):
        """
        Calculates the bounds of the variables used
        """
        super(Conv4, self)._calculate_bounds()

        time_steps = len(self.set_t_performance)

        # Output Bounds
        self.bounds["output"][
            self.input_parameters.performance_data["main_output_carrier"]
        ] = np.column_stack(
            (
                np.zeros(shape=(time_steps)),
                np.ones(shape=(time_steps)),
            )
        )

        for car in self.component_options.output_carrier:
            if not car == self.component_options.main_output_carrier:
                self.bounds["output"][car] = (
                    self.bounds["output"][self.component_options.main_output_carrier]
                    * self.input_parameters.performance_data["output_ratios"][car]
                )

    def construct_tech_model(self, b_tec, data: dict, set_t_full, set_t_clustered):
        """
        Adds constraints to technology blocks for tec_type CONV4

        :param b_tec: pyomo block with technology model
        :param dict data: data containing model configuration
        :param set_t_full: pyomo set containing timesteps
        :param set_t_clustered: pyomo set containing clustered timesteps
        :return: pyomo block with technology model
        """
        super(Conv4, self).construct_tech_model(
            b_tec, data, set_t_full, set_t_clustered
        )

        # DATA OF TECHNOLOGY
        coeff_ti = self.processed_coeff.time_independent
        rated_power = self.input_parameters.rated_power
        phi = coeff_ti["phi"]

        # add additional constraints for performance type 2 (min. part load)
        if self.component_options.performance_function_type == 2:
            b_tec = self._performance_function_type_2(b_tec)

        # Size constraints
        # constraint on output ratios
        def init_output_output(const, t, car_output):
            if car_output == self.component_options.main_output_carrier:
                return pyo.Constraint.Skip
            else:
                return (
                    self.output[t, car_output]
                    == phi[car_output]
                    * self.output[t, self.component_options.main_output_carrier]
                )

        b_tec.const_output_output = pyo.Constraint(
            self.set_t_performance, b_tec.set_output_carriers, rule=init_output_output
        )

        # size constraint based on main carrier output
        def init_size_constraint(const, t):
            return (
                self.output[t, self.component_options.main_output_carrier]
                <= b_tec.var_size * rated_power
            )

        b_tec.const_size = pyo.Constraint(
            self.set_t_performance, rule=init_size_constraint
        )

        return b_tec

    def _performance_function_type_2(self, b_tec):
        """
        Sets the minimum part load constraint for a tec based on tec_type CONV4 with
        performance type 2.

        :param b_tec: pyomo block with technology model
        :return: pyomo block with technology model
        """
        # Transformation required
        self.big_m_transformation_required = 1

        # Performance Parameters
        rated_power = self.input_parameters.rated_power
        coeff_ti = self.processed_coeff.time_independent
        min_part_load = coeff_ti["min_part_load"]

        # define disjuncts
        s_indicators = range(0, 2)

        def init_output(dis, t, ind):
            if ind == 0:  # technology off

                def init_output_off(const, car_output):
                    return self.output[t, car_output] == 0

                dis.const_output_off = pyo.Constraint(
                    b_tec.set_output_carriers, rule=init_output_off
                )

            else:  # technology on

                def init_min_partload(const):
                    return (
                        self.output[t, self.component_options.main_output_carrier]
                        >= min_part_load * b_tec.var_size * rated_power
                    )

                dis.const_min_partload = pyo.Constraint(rule=init_min_partload)

        b_tec.dis_output = gdp.Disjunct(
            self.set_t_performance, s_indicators, rule=init_output
        )

        # Bind disjuncts
        def bind_disjunctions(dis, t):
            return [b_tec.dis_output[t, i] for i in s_indicators]

        b_tec.disjunction_output = gdp.Disjunction(
            self.set_t_performance, rule=bind_disjunctions
        )

        return b_tec



import logging
from logging.handlers import MemoryHandler
from .modelhub import ModelHub as ModelHub
from .result_management import (
    print_h5_tree,
    extract_dataset_from_h5,
    extract_datasets_from_h5group,
)
from .diagnostics import get_infeasible_constraints
from .data_preprocessing import *
from .case_studies import *

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)

# Stream Handler to control console output
ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
logger.addHandler(ch)



from .construct_technology import construct_technology_block
from .construct_networks import construct_network_block
from .construct_balances import (
    construct_nodal_energybalance,
    construct_global_energybalance,
    construct_emission_balance,
    construct_system_cost,
    construct_network_constraints,
    delete_all_balances,
    construct_global_balance,
    construct_export_costs,
    construct_import_costs,
)
from .construct_nodes import construct_node_block
from .construct_investment_period import construct_investment_period_block
from .utilities import get_data_for_investment_period, get_data_for_node



from .utilities import get_attribute_from_dict


class ModelComponent:
    """
    Class to read and manage data for technologies and networks. This class inherits
    its attributes to the technology and network classes.
    """

    def __init__(self, data: dict):
        """
        Initializes component class

        Attributes include:

        - name: technology name
        - existing: if component is existing or not
        - size_initial: if existing, initial size
        - economics: contains economic data
        - input_parameters: unfitted parameters from json files
        - component_options: component options that are unrelated to the performance
          of the component
        - bounds: (for technologies only) containing bounds on input and output
           variables that are calculated in technology subclasses
        - processed_coeff: fitted/processed coefficients
        - big_m_transformation_required: flag to use for disjunctive programming

        :param dict data: technology/network data
        """
        self.name = data["name"]
        self.existing = 0
        self.size_initial = []
        self.economics = Economics(data["Economics"])

        self.input_parameters = InputParameters(data)
        self.component_options = ComponentOptions(data)
        self.bounds = {"input": {}, "output": {}}
        self.processed_coeff = ProcessedCoefficients()

        self.big_m_transformation_required = 0


class Economics:
    """
    Class to manage economic data of technologies and networks.

    Contains capex and opex data
    """

    def __init__(self, economics: dict):
        """
        Constructor

        :param dict economics: Dict containing economic data of component
        """
        if "CAPEX_model" in economics:
            self.capex_model = economics["CAPEX_model"]
        self.capex_data = {}
        if "unit_CAPEX" in economics:
            self.capex_data["unit_capex"] = economics["unit_CAPEX"]
        if "fix_CAPEX" in economics:
            self.capex_data["fix_capex"] = economics["fix_CAPEX"]
        if "piecewise_CAPEX" in economics:
            self.capex_data["piecewise_capex"] = economics["piecewise_CAPEX"]
        if "gamma1" in economics:
            self.capex_data["gamma1"] = economics["gamma1"]
            self.capex_data["gamma2"] = economics["gamma2"]
            self.capex_data["gamma3"] = economics["gamma3"]
            self.capex_data["gamma4"] = economics["gamma4"]
        self.opex_variable = economics["OPEX_variable"]
        self.opex_fixed = economics["OPEX_fixed"]
        self.discount_rate = economics["discount_rate"]
        self.lifetime = economics["lifetime"]
        self.decommission_cost = economics["decommission_cost"]


class InputParameters:
    """
    Class to hold unfitted/unprocessed performance of technologies and networks
    """

    def __init__(self, component_data: dict):
        self.performance_data = component_data["Performance"]
        self.size_min = component_data["size_min"]
        self.size_max = component_data["size_max"]
        self.rated_power = 1

        self.rated_power = get_attribute_from_dict(
            component_data["Performance"], "rated_power", 1
        )
        self.min_part_load = get_attribute_from_dict(
            component_data["Performance"], "min_part_load", 0
        )
        self.standby_power = get_attribute_from_dict(
            component_data["Performance"], "standby_power", -1
        )


class ComponentOptions:
    """
    Class to hold options for technologies and networks
    """

    def __init__(self, component_data: dict):
        self.modelled_with_full_res = False
        self.lower_res_than_full = False
        self.size_is_int = component_data["size_is_int"]
        self.decommission = component_data["decommission"]
        self.size_based_on = None
        self.emissions_based_on = None

        # TECHNOLOGY
        # Technology Model
        if "tec_type" in component_data:
            self.technology_model = component_data["tec_type"]

        # Input carrier
        self.input_carrier = get_attribute_from_dict(
            component_data["Performance"], "input_carrier", []
        )

        # Output Carriers
        self.output_carrier = get_attribute_from_dict(
            component_data["Performance"], "output_carrier", []
        )

        # Performance Function Type
        self.performance_function_type = get_attribute_from_dict(
            component_data["Performance"], "performance_function_type", None
        )

        # CCS
        if (
            "ccs" in component_data["Performance"]
            and component_data["Performance"]["ccs"]["possible"]
        ):
            self.ccs_possible = True
            self.ccs_type = component_data["Performance"]["ccs"]["ccs_type"]
        else:
            self.ccs_possible = False
            self.ccs_type = None

        # Standby power
        self.standby_power_carrier = get_attribute_from_dict(
            component_data["Performance"], "standby_power_carrier", -1
        )

        # Determined in child classes
        self.main_input_carrier = None
        self.main_output_carrier = None

        # NETWORKS
        # Transported carrier
        if "carrier" in component_data["Performance"]:
            self.transported_carrier = component_data["Performance"]["carrier"]

        if "energyconsumption" in component_data["Performance"]:
            if component_data["Performance"]["energyconsumption"]:
                self.energyconsumption = 1
            else:
                self.energyconsumption = 0

        # disable bidirectional for networks and storage
        if "bidirectional_network" in component_data["Performance"]:
            self.bidirectional_network = component_data["Performance"][
                "bidirectional_network"
            ]
            if self.bidirectional_network:
                self.bidirectional_network_precise = get_attribute_from_dict(
                    component_data["Performance"], "bidirectional_network_precise", 1
                )

        if "allow_only_one_direction" in component_data["Performance"]:
            self.allow_only_one_direction = component_data["Performance"][
                "allow_only_one_direction"
            ]
            if self.allow_only_one_direction:
                self.allow_only_one_direction_precise = get_attribute_from_dict(
                    component_data["Performance"], "allow_only_one_direction_precise", 1
                )
        # other technology specific options
        self.other = {}


class ProcessedCoefficients:
    """
    Defines a simple class for fitted/processed coefficients
    """

    def __init__(self):
        self.time_dependent_full = {}
        self.time_dependent_clustered = {}
        self.time_dependent_averaged = {}
        self.time_dependent_used = {}
        self.time_independent = {}
        self.dynamics = {}



import pyomo.environ as pyo
import pyomo.gdp as gdp
import copy
import numpy as np
import pandas as pd

from ..technology import Technology
from ...utilities import link_full_resolution_to_clustered


class GasTurbine(Technology):
    """
    Gas turbine

    Resembles gas turbines of different sizes.
    Hydrogen and Natural Gas Turbines are possible at four different sizes,
    as indicated by the file names of the data. Performance data and the model is
    taken from Weimann, L., Ellerker, M., Kramer, G. J., & Gazzani, M. (2019).
    Modeling gas turbines in multi-energy systems: A linear model accounting for
    part-load operation, fuel, temperature, and sizing effects. International
    Conference on Applied Energy. https://doi.org/10.46855/energy-proceedings-5280

    A small adaption is made: Natural gas turbines can co-fire hydrogen up to 5% of
    the energy content

    **Variable declarations:**

    - Total fuel input in :math:`t`: :math:`Input_{tot, t}`

    - Number of turbines on in :math:`t`: :math:`N_{on,t}`

    **Constraint declarations:**

    The following constants are used:

    - :math:`Input_{min}`: Minimal input per turbine

    - :math:`Input_{max}`: Maximal input per turbine

    - :math:`in_{H2max}`: Maximal H2 admixture to fuel (only for natural gas turbines, default is 0.05)

    - :math:`{\\alpha}`: Performance parameter for electricity output

    - :math:`{\\beta}`: Performance parameter for electricity output

    - :math:`{\\epsilon}`: Performance parameter for heat output

    - :math:`f({\\Theta})`: Ambient temperature correction factor

    - Input calculation (For hydrogen turbines, :math:`Input_{NG, t}` is zero, and the second constraint is removed):

      .. math::
        Input_{H2, t} + Input_{NG, t} = Input_{tot, t}

      .. math::
        Input_{H2, t} \\leq in_{H2max} Input_{tot, t}

    - Turbines on:

      .. math::
        N_{on, t} \\leq S

    - If technology is on:

      .. math::
        Output_{el,t} = ({\\alpha} Input_{tot, t} + {\\beta} * N_{on, t}) *f({\\Theta})

      .. math::
        Output_{th,t} = {\\epsilon} Input_{tot, t} - Output_{el,t}

      .. math::
        Input_{min} * N_{on, t} \\leq Input_{tot, t} \\leq Input_{max} * N_{on, t}

    - If the technology is off, input and output is set to 0:

      .. math::
         \\sum(Output_{t, car}) = 0

      .. math::
         \\sum(Input_{t, car}) = 0

    - Additionally, ramping rates of the technology can be constraint.

      .. math::
         -rampingrate \\leq \\sum(Input_{t, car}) - \\sum(Input_{t-1, car}) \\leq rampingrate

    """

    def __init__(self, tec_data: dict):
        """
        Constructor

        :param dict tec_data: technology data
        """
        super().__init__(tec_data)

        self.component_options.emissions_based_on = "input"
        self.component_options.size_based_on = "output"
        self.component_options.main_input_carrier = tec_data["Performance"][
            "main_input_carrier"
        ]

    def fit_technology_performance(self, climate_data: pd.DataFrame, location: dict):
        """
        Performs fitting for technology type GasTurbine

        :param tec_data: technology data
        :param climate_data: climate data
        :return:
        """
        super(GasTurbine, self).fit_technology_performance(climate_data, location)

        # Climate data & Number of timesteps
        time_steps = len(climate_data)

        # Ambient air temperature
        T = copy.deepcopy(climate_data["temp_air"])

        # Temperature correction factors
        f = np.empty(shape=(time_steps))
        f[T <= 6] = (
            self.input_parameters.performance_data["gamma"][0]
            * (T[T <= 6] / self.input_parameters.performance_data["T_iso"])
            + self.input_parameters.performance_data["delta"][0]
        )
        f[T > 6] = (
            self.input_parameters.performance_data["gamma"][1]
            * (T[T > 6] / self.input_parameters.performance_data["T_iso"])
            + self.input_parameters.performance_data["delta"][1]
        )

        # Derive return
        fit = {}
        fit["td"] = {}
        fit["td"]["temperature_correction"] = f.round(5)

        fit["ti"] = {}
        fit["ti"]["alpha"] = round(self.input_parameters.performance_data["alpha"], 5)
        fit["ti"]["beta"] = round(self.input_parameters.performance_data["beta"], 5)
        fit["ti"]["epsilon"] = round(
            self.input_parameters.performance_data["epsilon"], 5
        )
        fit["ti"]["in_min"] = round(self.input_parameters.performance_data["in_min"], 5)
        fit["ti"]["in_max"] = round(self.input_parameters.performance_data["in_max"], 5)
        if len(self.component_options.input_carrier) == 2:
            fit["ti"]["max_H2_admixture"] = self.input_parameters.performance_data[
                "max_H2_admixture"
            ]
        else:
            fit["ti"]["max_H2_admixture"] = 1

        # Coefficients
        for par in fit["td"]:
            self.processed_coeff.time_dependent_full[par] = fit["td"][par]
        for par in fit["ti"]:
            self.processed_coeff.time_independent[par] = fit["ti"][par]

    def _calculate_bounds(self):
        """
        Calculates the bounds of the variables used
        """
        super(GasTurbine, self)._calculate_bounds()

        time_steps = len(self.set_t_performance)

        bounds = {}

        # Input bounds
        bounds["input_bounds"] = {}
        for c in self.component_options.input_carrier:
            if c == "hydrogen":
                bounds["input_bounds"][c] = np.column_stack(
                    (
                        np.zeros(shape=(time_steps)),
                        np.ones(shape=(time_steps))
                        * self.input_parameters.performance_data["in_max"]
                        * self.processed_coeff.time_independent["max_H2_admixture"],
                    )
                )
            else:
                bounds["input_bounds"][c] = np.column_stack(
                    (
                        np.zeros(shape=(time_steps)),
                        np.ones(shape=(time_steps))
                        * self.input_parameters.performance_data["in_max"],
                    )
                )

        # Output bounds
        bounds["output_bounds"] = {}
        bounds["output_bounds"]["electricity"] = np.column_stack(
            (
                np.zeros(shape=(time_steps)),
                self.processed_coeff.time_dependent_used["temperature_correction"]
                * (
                    self.input_parameters.performance_data["in_max"]
                    * self.processed_coeff.time_independent["alpha"]
                    + self.processed_coeff.time_independent["beta"]
                ),
            )
        )
        bounds["output_bounds"]["heat"] = np.column_stack(
            (
                np.zeros(shape=(time_steps)),
                self.processed_coeff.time_independent["epsilon"]
                * self.processed_coeff.time_independent["in_max"]
                - self.processed_coeff.time_dependent_used["temperature_correction"]
                * (
                    self.input_parameters.performance_data["in_max"]
                    * self.processed_coeff.time_independent["alpha"]
                    + self.processed_coeff.time_independent["beta"]
                ),
            )
        )

        # Output Bounds
        self.bounds["output"] = bounds["output_bounds"]
        # Input Bounds
        for car in self.component_options.input_carrier:
            self.bounds["input"][car] = np.column_stack(
                (np.zeros(shape=(time_steps)), np.ones(shape=(time_steps)))
            )

    def construct_tech_model(self, b_tec, data: dict, set_t_full, set_t_clustered):
        """
        Adds constraints to technology blocks for gas turbines

        :param b_tec: pyomo block with technology model
        :param dict data: data containing model configuration
        :param set_t_full: pyomo set containing timesteps
        :param set_t_clustered: pyomo set containing clustered timesteps
        :return: pyomo block with technology model
        """
        super(GasTurbine, self).construct_tech_model(
            b_tec, data, set_t_full, set_t_clustered
        )

        # Transformation required
        self.big_m_transformation_required = 1

        # DATA OF TECHNOLOGY
        bounds = self.bounds
        coeff_td = self.processed_coeff.time_dependent_used
        coeff_ti = self.processed_coeff.time_independent
        dynamics = self.processed_coeff.dynamics

        # Parameter declaration
        in_min = coeff_ti["in_min"]
        in_max = coeff_ti["in_max"]
        max_H2_admixture = coeff_ti["max_H2_admixture"]
        alpha = coeff_ti["alpha"]
        beta = coeff_ti["beta"]
        epsilon = coeff_ti["epsilon"]
        temperature_correction = coeff_td["temperature_correction"]

        # Additional decision variables
        size_max = self.input_parameters.size_max

        def init_input_bounds(bds, t):
            if len(self.component_options.input_carrier) == 2:
                car = "gas"
            else:
                car = "hydrogen"
            return tuple(bounds["input"][car][t - 1, :] * size_max)

        b_tec.var_total_input = pyo.Var(
            self.set_t_performance,
            within=pyo.NonNegativeReals,
            bounds=init_input_bounds,
        )

        b_tec.var_units_on = pyo.Var(
            self.set_t_performance, within=pyo.NonNegativeIntegers, bounds=(0, size_max)
        )

        # Calculate total input
        def init_total_input(const, t):
            return b_tec.var_total_input[t] == sum(
                self.input[t, car_input] for car_input in b_tec.set_input_carriers
            )

        b_tec.const_total_input = pyo.Constraint(
            self.set_t_performance, rule=init_total_input
        )

        # Constrain hydrogen input
        if len(self.component_options.input_carrier) == 2:

            def init_h2_input(const, t):
                return (
                    self.input[t, "hydrogen"]
                    <= b_tec.var_total_input[t] * max_H2_admixture
                )

            b_tec.const_h2_input = pyo.Constraint(
                self.set_t_performance, rule=init_h2_input
            )

        # LINEAR, MINIMAL PARTLOAD
        s_indicators = range(0, 2)

        def init_input_output(dis, t, ind):
            if ind == 0:  # technology off

                def init_input_off(const, car):
                    return self.input[t, car] == 0

                dis.const_input = pyo.Constraint(
                    b_tec.set_input_carriers, rule=init_input_off
                )

                def init_output_off(const, car):
                    return self.output[t, car] == 0

                dis.const_output_off = pyo.Constraint(
                    b_tec.set_output_carriers, rule=init_output_off
                )

            else:  # technology on
                # input-output relation
                def init_input_output_on_el(const):
                    return (
                        self.output[t, "electricity"]
                        == (
                            alpha * b_tec.var_total_input[t]
                            + beta * b_tec.var_units_on[t]
                        )
                        * temperature_correction[t - 1]
                    )

                dis.const_input_output_on_el = pyo.Constraint(
                    rule=init_input_output_on_el
                )

                def init_input_output_on_th(const):
                    return (
                        self.output[t, "heat"]
                        == epsilon * b_tec.var_total_input[t]
                        - self.output[t, "electricity"]
                    )

                dis.const_input_output_on_th = pyo.Constraint(
                    rule=init_input_output_on_th
                )

                # min part load relation
                def init_min_input(const):
                    return b_tec.var_total_input[t] >= in_min * b_tec.var_units_on[t]

                dis.const_min_input = pyo.Constraint(rule=init_min_input)

                def init_max_input(const):
                    return b_tec.var_total_input[t] <= in_max * b_tec.var_units_on[t]

                dis.const_max_input = pyo.Constraint(rule=init_max_input)

        b_tec.dis_input_output = gdp.Disjunct(
            self.set_t_performance, s_indicators, rule=init_input_output
        )

        # Bind disjuncts
        def bind_disjunctions(dis, t):
            return [b_tec.dis_input_output[t, i] for i in s_indicators]

        b_tec.disjunction_input_output = gdp.Disjunction(
            self.set_t_performance, rule=bind_disjunctions
        )

        # Technologies on
        def init_n_on(const, t):
            return b_tec.var_units_on[t] <= b_tec.var_size

        b_tec.const_n_on = pyo.Constraint(self.set_t_performance, rule=init_n_on)

        # RAMPING RATES
        if "ramping_time" in dynamics:
            if not dynamics["ramping_time"] == -1:
                b_tec = self._define_ramping_rates(b_tec, data)

        return b_tec

    def write_results_tec_operation(self, h5_group, model_block):
        """
        Function to report results of technologies after optimization

        :param b_tec: technology model block
        :return: dict results: holds results
        """
        super(GasTurbine, self).write_results_tec_operation(h5_group, model_block)

        h5_group.create_dataset(
            "modules_on",
            data=[model_block.var_units_on[t].value for t in self.set_t_performance],
        )

    def _define_ramping_rates(self, b_tec, data):
        """
        Constraints the inputs for a ramping rate

        :param b_tec: technology model block
        :return:
        """
        dynamics = self.processed_coeff.dynamics

        ramping_time = dynamics["ramping_time"]

        # Calculate ramping rates
        if "ref_size" in dynamics and not dynamics["ref_size"] == -1:
            ramping_rate = dynamics["ref_size"] / ramping_time
        else:
            ramping_rate = b_tec.var_size / ramping_time

        # Constraints ramping rates
        if "ramping_const_int" in dynamics and dynamics["ramping_const_int"] == 1:

            s_indicators = range(0, 3)

            def init_ramping_operation_on(dis, t, ind):
                if t > 1:
                    if ind == 0:  # ramping constrained
                        dis.const_ramping_on = pyo.Constraint(
                            expr=b_tec.var_x[t] - b_tec.var_x[t - 1] == 0
                        )

                        def init_ramping_down_rate_operation(const):
                            return -ramping_rate <= sum(
                                self.input[t, car_input] - self.input[t - 1, car_input]
                                for car_input in b_tec.set_input_carriers
                            )

                        dis.const_ramping_down_rate = pyo.Constraint(
                            rule=init_ramping_down_rate_operation
                        )

                        def init_ramping_up_rate_operation(const):
                            return (
                                sum(
                                    self.input[t, car_input]
                                    - self.input[t - 1, car_input]
                                    for car_input in b_tec.set_input_carriers
                                )
                                <= ramping_rate
                            )

                        dis.const_ramping_up_rate = pyo.Constraint(
                            rule=init_ramping_up_rate_operation
                        )

                    elif ind == 1:  # startup, no ramping constraint
                        dis.const_ramping_on = pyo.Constraint(
                            expr=b_tec.var_x[t] - b_tec.var_x[t - 1] == 1
                        )

                    else:  # shutdown, no ramping constraint
                        dis.const_ramping_on = pyo.Constraint(
                            expr=b_tec.var_x[t] - b_tec.var_x[t - 1] == -1
                        )

            b_tec.dis_ramping_operation_on = gdp.Disjunct(
                self.set_t_performance, s_indicators, rule=init_ramping_operation_on
            )

            # Bind disjuncts
            def bind_disjunctions(dis, t):
                return [b_tec.dis_ramping_operation_on[t, i] for i in s_indicators]

            b_tec.disjunction_ramping_operation_on = gdp.Disjunction(
                self.set_t_performance, rule=bind_disjunctions
            )

        else:
            if data["config"]["optimization"]["typicaldays"]["N"]["value"] == 0:
                input_aux_rr = self.input
                set_t_rr = self.set_t_performance
            else:
                if (
                    data["config"]["optimization"]["typicaldays"]["method"]["value"]
                    == 1
                ):
                    sequence = data["k_means_specs"]["sequence"]
                elif (
                    data["config"]["optimization"]["typicaldays"]["method"]["value"]
                    == 2
                ):
                    sequence = self.sequence

                # init bounds at full res
                bounds_rr_full = {
                    "input": self.fitting_class.calculate_input_bounds(
                        self.component_options.size_based_on, len(self.set_t_full)
                    )
                }

                # create input variable for full res
                def init_input_bounds(bounds, t, car):
                    return tuple(
                        bounds_rr_full["input"][car][t - 1, :]
                        * self.processed_coeff.time_independent["size_max"]
                        * self.processed_coeff.time_independent["rated_power"]
                    )

                b_tec.var_input_rr_full = pyo.Var(
                    self.set_t_full,
                    b_tec.set_input_carriers,
                    within=pyo.NonNegativeReals,
                    bounds=init_input_bounds,
                )

                b_tec.const_link_full_resolution_rr = link_full_resolution_to_clustered(
                    self.input,
                    b_tec.var_input_rr_full,
                    self.set_t_full,
                    sequence,
                    b_tec.set_input_carriers,
                )

                input_aux_rr = b_tec.var_input_rr_full
                set_t_rr = self.set_t_full

            # Ramping constraint without integers
            def init_ramping_down_rate(const, t):
                if t > 1:
                    return -ramping_rate <= sum(
                        input_aux_rr[t, car_input] - input_aux_rr[t - 1, car_input]
                        for car_input in b_tec.set_input_carriers
                    )
                else:
                    return pyo.Constraint.Skip

            b_tec.const_ramping_down_rate = pyo.Constraint(
                set_t_rr, rule=init_ramping_down_rate
            )

            def init_ramping_up_rate(const, t):
                if t > 1:
                    return (
                        sum(
                            input_aux_rr[t, car_input] - input_aux_rr[t - 1, car_input]
                            for car_input in b_tec.set_input_carriers
                        )
                        <= ramping_rate
                    )
                else:
                    return pyo.Constraint.Skip

            b_tec.const_ramping_up_rate = pyo.Constraint(
                set_t_rr, rule=init_ramping_up_rate
            )

        return b_tec



import os
from pathlib import Path


def create_unique_folder_name(path: Path, name: str) -> Path:
    """
    Creates a unique folder name, in case the specified name already exists in the given path.

    The unique folder name is either the given folder name if the folder name did not exist yet in the given path, or
    the given folder name with an added suffix (_1, _2, _3, etc.).
    :param Path path: path to check
    :param str name: folder name
    :return: path to the folder with the unique folder name
    :rtype: Path
    """
    folder_path = Path.joinpath(path, name)
    counter = 1
    while folder_path.is_dir():
        folder_path = folder_path.with_name(
            f"{folder_path.stem}_{counter}{folder_path.suffix}"
        )
        counter += 1
    return folder_path


def create_save_folder(save_path: Path):
    """
    Creates a new folder at save_path

    :param Path save_path: path at which the folder is created
    :return:
    """
    os.makedirs(save_path)



import warnings

import pyomo.environ as pyo
import pyomo.gdp as gdp
import copy
import numpy as np
import pandas as pd
from pathlib import Path
from scipy.interpolate import griddata

from ..technology import Technology
from ...utilities import link_full_resolution_to_clustered
from ..utilities import fit_piecewise_function


import logging

log = logging.getLogger(__name__)


class CCPP(Technology):
    """
    Combined Cycle Power Plant with Steam Production

    Resembles Combined Cycle Power Plant with Steam Production.
    The size of the power plant it fixed and it only possible to model the plant as
    an existing technology (i.e. it cannot be sized). The model is based on Wiegner
    et al. (2024). Optimizing the Use of Limited Amounts of Hydrogen in Existing
    Combined Heat and Power Plants
    https://www.sciencedirect.com/science/article/pii/S2667095X24000199.

    It is possible in the JSON file to specify the following options:

    - size_ohb: Size of an Oxy-fuel hydrogen burner

    - size_db: Size of a duct burner that can burn hydrogen or natural gas

    - max_h2_in_gt: Max share (energy-based) of hydrogen combustion in the gas
      turbine ("max_input")

    - max_h2_in: Maximum of hydrogen input to plant

    - How to treat steam production ("steam_production"):

        - "ignore": technology has no additional output other than electricity

        - "as heat": technology has heat output, equivalent to the sum of HP/MP
          generation

        - "as steam": technology has steam output, equivalent to the sum of HP/MP
          generation

        - "as hp/mp steam": technology has HP and MP steam output

    - What additional components in the plant should be considered. Possible options
      are:

        - "None": only CCPP

        - "DB": Duct burner

        - "OHB": Oxyfuel-hydrogen burner

    - Number of segments to use for performance function ("nr_segments").
      Recommended: 2

    **Variable declarations:**

    - Total fuel input in :math:`t`: :math:`Input_{tot, t}`

    - Fuel input to duct burner/ OHB :math:`Input_{OHB/DB, t}`

    **Constraint declarations:**

    The following constants are used:

    - :math:`Input_{min}`: Minimal input per turbine

    - :math:`Input_{max}`: Maximal input per turbine

    - :math:`in_{H2max}`: Maximal H2 admixture to fuel (only for natural gas turbines, default is 0.05)

    - :math:`{\\alpha}`: Performance parameter for electricity output

    - :math:`{\\beta}`: Performance parameter for electricity output

    - :math:`{\\epsilon}`: Performance parameter for heat output

    - :math:`f({\\Theta})`: Ambient temperature correction factor

    - Input calculation (For hydrogen turbines, :math:`Input_{NG, t}` is zero, and the second constraint is removed):

      .. math::
        Input_{H2, t} + Input_{NG, t} = Input_{tot, t}

      .. math::
        Input_{H2, t} \\leq in_{H2max} Input_{tot, t}

    - Turbines on:

      .. math::
        N_{on, t} \\leq S

    - If technology is on:

      .. math::
        Output_{el,t} = ({\\alpha} Input_{tot, t} + {\\beta} * N_{on, t}) *f({\\Theta})

      .. math::
        Output_{th,t} = {\\epsilon} Input_{tot, t} - Output_{el,t}

      .. math::
        Input_{min} * N_{on, t} \\leq Input_{tot, t} \\leq Input_{max} * N_{on, t}

    - If the technology is off, input and output is set to 0:

      .. math::
        \\sum(Output_{t, car}) = 0

      .. math::
        \\sum(Input_{t, car}) = 0

    - Additionally, ramping rates of the technology can be constraint.

      .. math::
         -rampingrate \\leq\\sum(Input_{t, car}) -\\sum(Input_{t-1, car}) \\leq rampingrate

    """

    def __init__(self, tec_data: dict):
        """
        Constructor

        :param dict tec_data: technology data
        """
        super().__init__(tec_data)

        self.component_options.emissions_based_on = "input"
        self.component_options.size_based_on = "output"
        self.component_options.main_input_carrier = tec_data["Performance"][
            "main_input_carrier"
        ]

        # Treatment of Steam Production
        self.component_options.other["steam_production"] = tec_data["Performance"][
            "steam_production"
        ]
        steam_production = self.component_options.other["steam_production"]

        if steam_production == "ignore":
            self.component_options.output_carrier = ["electricity"]
            log_msg = "CCPP setting: only electricity output"

        elif steam_production == "as heat":
            self.component_options.output_carrier = ["electricity", "heat"]
            log_msg = "CCPP setting: only electricity, heat output"

        elif steam_production == "as steam":
            self.component_options.output_carrier = ["electricity", "steam"]
            log_msg = "CCPP setting: only electricity, steam output"

        elif steam_production == "as hp/mp steam":
            self.component_options.output_carrier = [
                "electricity",
                "steam_hp",
                "steam_mp",
            ]
            log_msg = "CCPP setting: electricity, steam_hp, steam_mp output"

        else:
            raise Exception("steam_production setting in CCPP incorrectly specified")
        log.info(log_msg)

        # Addtional Components considered
        self.component_options.other["component"] = tec_data["Performance"]["component"]

        # Number of segments
        self.component_options.other["nr_segments"] = tec_data["Performance"][
            "nr_segments"
        ]

        # All other options
        self.component_options.other["steam_turbine_generator_efficiency"] = tec_data[
            "Performance"
        ]["steam_turbine_generator_efficiency"]
        self.component_options.other["max_steam_extract_HP"] = 140
        self.component_options.other["max_steam_extract_MP"] = 51
        self.component_options.other["kappa_steam"] = 0.45

    def fit_technology_performance(self, climate_data: pd.DataFrame, location: dict):
        """
        Performs fitting for technology type CCPP

        :param tec_data: technology data
        :param climate_data: climate data
        :return:
        """
        super(CCPP, self).fit_technology_performance(climate_data, location)

        # Climate data & Number of timesteps
        T = copy.deepcopy(climate_data["temp_air"])

        # Remove outliers
        T[T >= 30] = 30
        T[T <= 0] = 0

        # Determine correct reading paths
        data_path = {}
        if "performance_data_path" in self.input_parameters.performance_data:
            performance_data_path = self.input_parameters.performance_data[
                "performance_data_path"
            ]
        else:
            performance_data_path = Path(__file__).parent.parent.parent.parent
            performance_data_path = (
                performance_data_path
                / "data/technology_data/PowerGeneration/CombinedCycle_fixed_size_data"
            )
            log.warning(
                "Using performance data with noice (not the same as in "
                "publication. Refer to the paper authors (Jan Wiegner) to get "
                "the actual data."
            )

        data_path["GT"] = Path("GT_fitting_data.csv")
        data_path["HP"] = Path("HP_fitting_data.csv")
        data_path["MP"] = Path("MP_fitting_data.csv")

        # Fit GT performance
        perf_data = {}
        perf_data["GT"] = pd.read_csv(performance_data_path / data_path["GT"], sep=";")

        for turbine in ["HP", "MP"]:
            perf_data[turbine] = pd.read_csv(
                performance_data_path / data_path[turbine],
                sep=";",
                index_col=0,
                header=[0, 1],
            )
            perf_data[turbine] = pd.read_csv(
                performance_data_path / data_path[turbine],
                sep=";",
                index_col=0,
                header=[0, 1],
            )

        # Fit to temperature
        igv_positions = list(perf_data["GT"]["IGV"].unique())
        nr_igv_positions = len(igv_positions)
        gt_eta_el = np.empty(shape=(len(T), nr_igv_positions))
        gt_p_in = np.empty(shape=(len(T), nr_igv_positions))
        gt_alpha_th = np.empty(shape=(len(T), nr_igv_positions))

        pos = 0
        for igv_position in igv_positions:
            data = perf_data["GT"][perf_data["GT"]["IGV"] == igv_position]
            gt_eta_el[:, pos] = griddata(
                np.array(data["T"]), np.array(data["GT_eta"]), T, method="linear"
            )
            gt_p_in[:, pos] = griddata(
                np.array(data["T"]), np.array(data["GT_P_in"]), T, method="linear"
            )
            data["share_thermal"] = (data["GT_P_el"] + data["GT_P_th"]) / data[
                "GT_P_in"
            ]
            gt_alpha_th[:, pos] = griddata(
                np.array(data["T"]), np.array(data["share_thermal"]), T, method="linear"
            )
            pos += 1

        # Fit performance (piecewise performance GT)
        nr_segments = self.component_options.other["nr_segments"]

        alpha_el = np.empty(shape=(len(T), nr_segments))
        beta_el = np.empty(shape=(len(T), nr_segments))
        bp_el_x = np.empty(shape=(len(T), nr_segments + 1))
        bp_el_y = np.empty(shape=(len(T), nr_segments + 1))
        alpha_th = gt_alpha_th.mean(axis=1)

        log.info("Deriving GT performance for CCPP...")

        for timestep in range(len(T)):
            if timestep % 100 == 1:
                print("\rComplete: ", round(timestep / len(T), 2) * 100, "%", end="")

            # Input-Output relation
            y = {}
            y["out_el"] = gt_p_in[timestep, :] * gt_eta_el[timestep, :]
            time_step_fit = fit_piecewise_function(
                gt_p_in[timestep, :], y, int(nr_segments)
            )
            alpha_el[timestep, :] = time_step_fit["out_el"]["alpha1"]
            beta_el[timestep, :] = time_step_fit["out_el"]["alpha2"]
            bp_el_x[timestep, :] = time_step_fit["out_el"]["bp_x"]
            bp_el_y[timestep, :] = time_step_fit["out_el"]["bp_y"]

        print("Complete: ", 100, "%")

        self.processed_coeff.time_dependent_full["GT"] = {}
        self.processed_coeff.time_dependent_full["GT"]["alpha_el"] = alpha_el
        self.processed_coeff.time_dependent_full["GT"]["beta_el"] = beta_el
        self.processed_coeff.time_dependent_full["GT"]["bp_el_x"] = bp_el_x
        self.processed_coeff.time_dependent_full["GT"]["bp_el_y"] = bp_el_y
        self.processed_coeff.time_dependent_full["GT"]["alpha_th"] = alpha_th

        # Fit HP/MP performance
        bp = {}
        bp["HP"] = [0, 12.92, 100]
        bp["MP"] = [0, 42.50, 100]
        for p in ["HP", "MP"]:

            alpha_gt = np.empty(shape=(len(T), nr_segments))
            alpha_hp = np.empty(shape=(len(T), nr_segments))
            alpha_mp = np.empty(shape=(len(T), nr_segments))
            alpha_cst = np.empty(shape=(len(T), nr_segments))
            if self.component_options.other["component"] == "DB":
                alpha_db = np.empty(shape=(len(T), nr_segments))
            elif self.component_options.other["component"] == "OHB":
                alpha_ohb = np.empty(shape=(len(T), nr_segments))

            for par in [1, 2]:
                data = perf_data[p]["GT"]["alpha_" + str(par)]
                alpha_gt[:, par - 1] = griddata(
                    np.array(data.index), np.array(data), T, method="linear"
                )

                data = perf_data[p]["HP"]["alpha_" + str(par)]
                alpha_hp[:, par - 1] = griddata(
                    np.array(data.index), np.array(data), T, method="linear"
                )

                data = perf_data[p]["MP"]["alpha_" + str(par)]
                alpha_mp[:, par - 1] = griddata(
                    np.array(data.index), np.array(data), T, method="linear"
                )

                data = perf_data[p]["cst"]["alpha_" + str(par)]
                alpha_cst[:, par - 1] = griddata(
                    np.array(data.index), np.array(data), T, method="linear"
                )

                if self.component_options.other["component"] == "DB":
                    data = perf_data[p]["DB"]["alpha_" + str(par)]
                    alpha_db[:, par - 1] = griddata(
                        np.array(data.index), np.array(data), T, method="linear"
                    )

                elif self.component_options.other["component"] == "OHB":
                    data = perf_data[p]["OHB"]["alpha_" + str(par)]
                    alpha_ohb[:, par - 1] = griddata(
                        np.array(data.index), np.array(data), T, method="linear"
                    )

            self.processed_coeff.time_independent[p] = {}
            self.processed_coeff.time_independent[p]["alpha_gt"] = alpha_gt
            self.processed_coeff.time_independent[p]["alpha_hp"] = alpha_hp
            self.processed_coeff.time_independent[p]["alpha_mp"] = alpha_mp
            self.processed_coeff.time_independent[p]["alpha_cst"] = alpha_cst
            self.processed_coeff.time_independent[p]["bp_" + p.lower()] = bp[p]
            if self.component_options.other["component"] == "DB":
                self.processed_coeff.time_independent[p]["alpha_db"] = alpha_db
            elif self.component_options.other["component"] == "OHB":
                self.processed_coeff.time_independent[p]["alpha_ohb"] = alpha_ohb

        data = self.input_parameters.performance_data
        self.processed_coeff.time_independent["eta_stg"] = data[
            "steam_turbine_generator_efficiency"
        ]

        self.processed_coeff.time_independent["size_db"] = data["size_db"]
        self.processed_coeff.time_independent["size_ohb"] = data["size_ohb"]
        self.processed_coeff.time_independent["max_h2_in"] = data["max_h2_in"]
        self.processed_coeff.time_independent["max_h2_in_gt"] = data["max_h2_in_gt"]
        self.processed_coeff.time_independent["hp_max_p"] = 27
        self.processed_coeff.time_independent["hp_min_p"] = 3
        self.processed_coeff.time_independent["mp_max_p"] = 90
        self.processed_coeff.time_independent["mp_min_p"] = 12

    def _calculate_bounds(self):
        """
        Calculates the bounds of the variables used
        """
        super(CCPP, self)._calculate_bounds()

        time_steps = len(self.set_t_performance)

        # Input bounds
        min_in = np.zeros(shape=(time_steps))

        # H2
        max_h2 = (
            np.ones(shape=(time_steps))
            * self.processed_coeff.time_independent["max_h2_in"]
        )

        # NG
        if self.component_options.other["component"] == "DB":
            additional_ng_in = self.processed_coeff.time_independent["size_db"]
        else:
            additional_ng_in = 0
        max_ng = (
            self.processed_coeff.time_dependent_full["GT"]["bp_el_x"][:, -1]
            + additional_ng_in
        )

        self.bounds["input"]["hydrogen"] = np.column_stack((min_in, max_h2))
        self.bounds["input"]["gas"] = np.column_stack((min_in, max_ng))
        self.bounds["input"]["total"] = np.column_stack((min_in, max_ng + max_h2))

        # Output bounds
        min_out = np.zeros(shape=(time_steps))
        max_el = (
            self.processed_coeff.time_dependent_full["GT"]["bp_el_y"][:, -1]
            + self.processed_coeff.time_independent["hp_max_p"]
            + self.processed_coeff.time_independent["mp_max_p"]
        )
        max_hp = (
            np.ones(shape=(time_steps))
            * self.component_options.other["max_steam_extract_HP"]
        )
        max_mp = (
            np.ones(shape=(time_steps))
            * self.component_options.other["max_steam_extract_MP"]
        )
        self.bounds["output"]["electricity"] = np.column_stack((min_out, max_el))

        if self.component_options.other["steam_production"] == "ignore":
            # For completeness, no other output then electricity
            pass

        elif self.component_options.other["steam_production"] == "as heat":
            self.bounds["output"]["heat"] = np.column_stack((min_out, max_hp + max_mp))

        elif self.component_options.other["steam_production"] == "as steam":
            self.bounds["output"]["steam"] = np.column_stack((min_out, max_hp + max_mp))

        elif self.component_options.other["steam_production"] == "as hp/mp steam":
            self.bounds["output"]["steam_hp"] = np.column_stack((min_out, max_hp))
            self.bounds["output"]["steam_mp"] = np.column_stack((min_out, max_mp))

    def construct_tech_model(self, b_tec, data: dict, set_t_full, set_t_clustered):
        """
        Adds constraints to technology blocks for gas turbines

        :param b_tec: pyomo block with technology model
        :param dict data: data containing model configuration
        :param set_t_full: pyomo set containing timesteps
        :param set_t_clustered: pyomo set containing clustered timesteps
        :return: pyomo block with technology model
        """
        super(CCPP, self).construct_tech_model(b_tec, data, set_t_full, set_t_clustered)

        b_tec = self._define_additional_vars(b_tec)
        b_tec = self._define_tec_global_balances(b_tec)
        b_tec = self._define_performance(b_tec)

        self.big_m_transformation_required = 1

        # RAMPING RATES
        dynamics = self.processed_coeff.dynamics
        if "ramping_time" in dynamics:
            if not dynamics["ramping_time"] == -1:
                b_tec = self._define_ramping_rates(b_tec, data)

        return b_tec

    def _define_additional_vars(self, b_tec):

        # DATA OF TECHNOLOGY
        coeff_td = self.processed_coeff.time_dependent_used
        coeff_ti = self.processed_coeff.time_independent

        # GT - total fuel input
        def init_input_total_gt_bounds(bds, t):
            return tuple((0, coeff_td["GT"]["bp_el_x"][t - 1, -1]))

        b_tec.var_gt_input = pyo.Var(
            self.set_t_performance,
            within=pyo.NonNegativeReals,
            bounds=init_input_total_gt_bounds,
        )

        # GT - ng fuel input
        def init_input_ng_gt_bounds(bds, t):
            return tuple((0, coeff_td["GT"]["bp_el_x"][t - 1, -1]))

        b_tec.var_gt_ng_input = pyo.Var(
            self.set_t_performance,
            within=pyo.NonNegativeReals,
            bounds=init_input_ng_gt_bounds,
        )

        # GT - h2 fuel input
        def init_input_h2_gt_bounds(bds, t):
            min_h2 = min(
                coeff_td["GT"]["bp_el_x"][t - 1, -1] * coeff_ti["max_h2_in_gt"],
                coeff_ti["max_h2_in"],
            )
            return tuple((0, min_h2))

        b_tec.var_gt_h2_input = pyo.Var(
            self.set_t_performance,
            within=pyo.NonNegativeReals,
            bounds=init_input_h2_gt_bounds,
        )

        # GT - P electric
        def init_gt_p_el_bounds(bds, t):
            return tuple((0, coeff_td["GT"]["bp_el_y"][t - 1, -1]))

        b_tec.var_gt_p_el = pyo.Var(
            self.set_t_performance,
            within=pyo.NonNegativeReals,
            bounds=init_gt_p_el_bounds,
        )

        # GT - P thermal
        gt_alpha_th = coeff_td["GT"]["alpha_th"]
        gt_bp_y = coeff_td["GT"]["bp_el_y"]
        gt_bp_x = coeff_td["GT"]["bp_el_x"]

        def init_gt_p_th_bounds(bds, t):
            return tuple(
                (0, gt_alpha_th[t - 1] * (gt_bp_x[t - 1, -1] - gt_bp_y[t - 1, -1]))
            )

        b_tec.var_gt_p_th = pyo.Var(
            self.set_t_performance,
            within=pyo.NonNegativeReals,
            bounds=init_gt_p_th_bounds,
        )

        # Duct Burner - input
        if self.component_options.other["component"] == "DB":

            def init_db_input_bounds(bds, t):
                return tuple((0, coeff_ti["size_db"]))

            b_tec.var_db_h2_input = pyo.Var(
                self.set_t_performance,
                within=pyo.NonNegativeReals,
                bounds=init_db_input_bounds,
            )
            b_tec.var_db_ng_input = pyo.Var(
                self.set_t_performance,
                within=pyo.NonNegativeReals,
                bounds=init_db_input_bounds,
            )
            b_tec.var_db_input = pyo.Var(
                self.set_t_performance,
                within=pyo.NonNegativeReals,
                bounds=init_db_input_bounds,
            )
        else:
            b_tec.var_db_h2_input = pyo.Param(
                self.set_t_performance, domain=pyo.Reals, initialize=0
            )
            b_tec.var_db_ng_input = pyo.Param(
                self.set_t_performance, domain=pyo.Reals, initialize=0
            )
            b_tec.var_db_input = pyo.Param(
                self.set_t_performance, domain=pyo.Reals, initialize=0
            )

        # OHB - input
        if self.component_options.other["component"] == "OHB":

            def init_ohb_input_bounds(bds, t):
                return tuple((0, coeff_ti["size_ohb"]))

            b_tec.var_ohb_h2_input = pyo.Var(
                self.set_t_performance,
                within=pyo.NonNegativeReals,
                bounds=init_ohb_input_bounds,
            )
        else:
            b_tec.var_ohb_h2_input = pyo.Param(
                self.set_t_performance, domain=pyo.Reals, initialize=0
            )

        # STHP electric output
        def init_hp_output_bounds(bds, t):
            return tuple((0, coeff_ti["hp_max_p"]))

        b_tec.var_hp_p_el = pyo.Var(
            self.set_t_performance,
            within=pyo.NonNegativeReals,
            bounds=init_hp_output_bounds,
        )

        # STMP electric output
        def init_mp_output_bounds(bds, t):
            return tuple((0, coeff_ti["mp_max_p"]))

        b_tec.var_mp_p_el = pyo.Var(
            self.set_t_performance,
            within=pyo.NonNegativeReals,
            bounds=init_mp_output_bounds,
        )

        # Steam Output
        def init_mp_hp_bounds(bds, t):
            return tuple((0, self.component_options.other["max_steam_extract_HP"]))

        b_tec.var_mp_p_hp = pyo.Var(
            self.set_t_performance,
            within=pyo.NonNegativeReals,
            bounds=init_mp_hp_bounds,
        )

        def init_mp_mp_bounds(bds, t):
            return tuple((0, self.component_options.other["max_steam_extract_MP"]))

        b_tec.var_mp_p_mp = pyo.Var(
            self.set_t_performance,
            within=pyo.NonNegativeReals,
            bounds=init_mp_mp_bounds,
        )

        def init_hp_hp_bounds(bds, t):
            return tuple((0, coeff_ti["hp_max_p"]))

        b_tec.var_hp_p_hp = pyo.Var(
            self.set_t_performance,
            within=pyo.NonNegativeReals,
            bounds=init_hp_hp_bounds,
        )

        def init_hp_mp_bounds(bds, t):
            return tuple((0, coeff_ti["mp_max_p"]))

        b_tec.var_hp_p_mp = pyo.Var(
            self.set_t_performance,
            within=pyo.NonNegativeReals,
            bounds=init_hp_mp_bounds,
        )

        if not b_tec.find_component("var_x"):
            b_tec.var_x = pyo.Var(
                self.set_t_performance, domain=pyo.NonNegativeIntegers, bounds=(0, 1)
            )

        return b_tec

    def _define_tec_global_balances(self, b_tec):

        # DATA OF TECHNOLOGY
        coeff_ti = self.processed_coeff.time_independent

        # Calculate total hydrogen input
        def init_total_input_h2(const, t):
            return (
                b_tec.var_input[t, "hydrogen"]
                == b_tec.var_gt_h2_input[t]
                + b_tec.var_db_h2_input[t]
                + b_tec.var_ohb_h2_input[t]
            )

        b_tec.const_total_input_h2 = pyo.Constraint(
            self.set_t_performance,
            rule=init_total_input_h2,
        )

        # Calculate total gas input
        def init_total_input_ng(const, t):
            return (
                b_tec.var_input[t, "gas"]
                == b_tec.var_gt_ng_input[t] + b_tec.var_db_ng_input[t]
            )

        b_tec.const_total_input_ng = pyo.Constraint(
            self.set_t_performance,
            rule=init_total_input_ng,
        )

        # Calculate total electric output
        def init_total_output_el(const, t):
            return b_tec.var_output[t, "electricity"] == b_tec.var_gt_p_el[
                t
            ] + coeff_ti["eta_stg"] * (b_tec.var_mp_p_el[t] + b_tec.var_hp_p_el[t])

        b_tec.const_total_output_el = pyo.Constraint(
            self.set_t_performance,
            rule=init_total_output_el,
        )

        if self.component_options.other["steam_production"] == "ignore":
            # For completeness, no other output than electricity
            pass

        elif self.component_options.other["steam_production"] == "as heat":

            def init_total_output_heat(const, t):
                return (
                    b_tec.var_output[t, "heat"]
                    == b_tec.var_mp_p_hp[t]
                    + b_tec.var_mp_p_mp[t]
                    + b_tec.var_hp_p_hp[t]
                    + b_tec.var_hp_p_mp[t]
                )

            b_tec.const_total_output_heat = pyo.Constraint(
                self.set_t_performance,
                rule=init_total_output_heat,
            )

        elif self.component_options.other["steam_production"] == "as steam":

            def init_total_output_steam(const, t):
                return (
                    b_tec.var_output[t, "steam"]
                    == b_tec.var_mp_p_hp[t]
                    + b_tec.var_mp_p_mp[t]
                    + b_tec.var_hp_p_hp[t]
                    + b_tec.var_hp_p_mp[t]
                )

            b_tec.const_total_output_steam = pyo.Constraint(
                self.set_t_performance,
                rule=init_total_output_steam,
            )

        elif self.component_options.other["steam_production"] == "as hp/mp steam":

            def init_total_output_steam_hp(const, t):
                return (
                    b_tec.var_output[t, "steam_hp"]
                    == b_tec.var_mp_p_hp[t] + b_tec.var_hp_p_hp[t]
                )

            b_tec.const_total_output_steam_hp = pyo.Constraint(
                self.set_t_performance,
                rule=init_total_output_steam_hp,
            )

            def init_total_output_steam_mp(const, t):
                return (
                    b_tec.var_output[t, "steam_mp"]
                    == b_tec.var_mp_p_mp[t] + b_tec.var_hp_p_mp[t]
                )

            b_tec.const_total_output_steam_mp = pyo.Constraint(
                self.set_t_performance,
                rule=init_total_output_steam_mp,
            )

        # Constrain total H2 input
        def init_total_input_h2_max(const, t):
            return b_tec.var_input[t, "hydrogen"] <= coeff_ti["max_h2_in"]

        b_tec.const_total_input_h2_max = pyo.Constraint(
            self.set_t_performance,
            rule=init_total_input_h2_max,
        )

        return b_tec

    def _define_performance(self, b_tec):
        coeff_td = self.processed_coeff.time_dependent_used
        coeff_ti = self.processed_coeff.time_independent
        nr_segments = self.component_options.other["nr_segments"]
        gt_alpha_th = coeff_td["GT"]["alpha_th"]
        gt_alpha = coeff_td["GT"]["alpha_el"]
        gt_beta = coeff_td["GT"]["beta_el"]
        gt_bp_x = coeff_td["GT"]["bp_el_x"]

        # Total input to GT
        def init_total_input_gt(const, t):
            return (
                b_tec.var_gt_input[t]
                == b_tec.var_gt_ng_input[t] + b_tec.var_gt_h2_input[t]
            )

        b_tec.const_total_input_gt = pyo.Constraint(
            self.set_t_performance,
            rule=init_total_input_gt,
        )

        # Max H2 in
        def init_max_h2_input_gt(const, t):
            return (
                b_tec.var_gt_h2_input[t]
                <= coeff_ti["max_h2_in_gt"] * b_tec.var_gt_input[t]
            )

        b_tec.const_max_h2_input_gt = pyo.Constraint(
            self.set_t_performance,
            rule=init_max_h2_input_gt,
        )

        s_indicators = range(0, nr_segments + 1)

        def init_input_output(dis, t, ind):
            if ind == 0:  # technology off

                dis.const_x_off = pyo.Constraint(expr=b_tec.var_x[t] == 0)

                def init_input_gt_ng_off(const):
                    return b_tec.var_gt_ng_input[t] == 0

                dis.const_input_gt_ng_off = pyo.Constraint(rule=init_input_gt_ng_off)

                def init_input_gt_h2_off(const):
                    return b_tec.var_gt_h2_input[t] == 0

                dis.const_input_gt_h2_off = pyo.Constraint(rule=init_input_gt_h2_off)

                def init_output_gt_el_off(const):
                    return b_tec.var_gt_p_el[t] == 0

                dis.const_output_gt_el_off = pyo.Constraint(rule=init_output_gt_el_off)

                def init_output_mp_off(const):
                    return b_tec.var_mp_p_el[t] == 0

                dis.const_output_mp_off = pyo.Constraint(rule=init_output_mp_off)

                def init_output_hp_off(const):
                    return b_tec.var_hp_p_el[t] == 0

                dis.const_output_hp_off = pyo.Constraint(rule=init_output_hp_off)

                def init_input_off(const, car_input):
                    return self.input[t, car_input] == 0

                dis.const_input = pyo.Constraint(
                    b_tec.set_input_carriers, rule=init_input_off
                )

                def init_output_off(const, car_output):
                    return self.output[t, car_output] == 0

                dis.const_output_off = pyo.Constraint(
                    b_tec.set_output_carriers, rule=init_output_off
                )

                if self.component_options.other["component"] == "DB":

                    def init_input_db_off(const):
                        return b_tec.var_db_input[t] == 0

                    dis.const_input_db_off = pyo.Constraint(rule=init_input_db_off)

                    def init_input_db_ng_off(const):
                        return b_tec.var_db_ng_input[t] == 0

                    dis.const_input_db_ng_off = pyo.Constraint(
                        rule=init_input_db_ng_off
                    )

                    def init_input_db_h2_off(const):
                        return b_tec.var_db_h2_input[t] == 0

                    dis.const_input_db_h2_off = pyo.Constraint(
                        rule=init_input_db_h2_off
                    )

                if self.component_options.other["component"] == "OHB":

                    def init_input_ohb_h2_off(const):
                        return b_tec.var_ohb_h2_input[t] == 0

                    dis.const_input_ohb_h2_off = pyo.Constraint(
                        rule=init_input_ohb_h2_off
                    )

            else:  # piecewise definition

                dis.const_x_on = pyo.Constraint(expr=b_tec.var_x[t] == 1)

                def init_input_on1(const):
                    return gt_bp_x[t - 1, ind - 1] <= b_tec.var_gt_input[t]

                dis.const_input_on1 = pyo.Constraint(rule=init_input_on1)

                def init_input_on2(const):
                    return b_tec.var_gt_input[t] <= gt_bp_x[t - 1, ind]

                dis.const_input_on2 = pyo.Constraint(rule=init_input_on2)

                def init_output_gt_on(const):
                    return (
                        b_tec.var_gt_p_el[t]
                        == gt_alpha[t - 1, ind - 1] * b_tec.var_gt_input[t]
                        + gt_beta[t - 1, ind - 1]
                    )

                dis.const_input_output_gt_on = pyo.Constraint(rule=init_output_gt_on)

        b_tec.dis_input_output = gdp.Disjunct(
            self.set_t_performance, s_indicators, rule=init_input_output
        )

        # Bind disjuncts
        def bind_disjunctions(dis, t):
            return [b_tec.dis_input_output[t, i] for i in s_indicators]

        b_tec.disjunction_input_output = gdp.Disjunction(
            self.set_t_performance, rule=bind_disjunctions
        )

        def init_output_gt_th(const, t):
            return b_tec.var_gt_p_th[t] == gt_alpha_th[t - 1] * (
                b_tec.var_gt_input[t] - b_tec.var_gt_p_el[t]
            )

        b_tec.const_output_gt_th = pyo.Constraint(
            self.set_t_performance, rule=init_output_gt_th
        )

        st_hp_alpha_gt = coeff_ti["HP"]["alpha_gt"]
        st_hp_alpha_hp = coeff_ti["HP"]["alpha_hp"]
        st_hp_alpha_mp = coeff_ti["HP"]["alpha_mp"]
        st_hp_alpha_cst = coeff_ti["HP"]["alpha_cst"]
        st_hp_bp_x = coeff_ti["HP"]["bp_hp"]
        if self.component_options.other["component"] == "DB":
            st_hp_alpha_db = coeff_ti["HP"]["alpha_db"]
        elif self.component_options.other["component"] == "OHB":
            st_hp_alpha_ohb = coeff_ti["HP"]["alpha_ohb"]
        s_indicators = range(0, nr_segments + 1)

        def init_perf_hp(dis, t, ind):
            if ind == 0:  # technology off
                dis.const_x_off = pyo.Constraint(expr=b_tec.var_x[t] == 0)

            else:  # technology on
                dis.const_x_on = pyo.Constraint(expr=b_tec.var_x[t] == 1)

                # HP bounds
                def init_hp_on1(const):
                    return st_hp_bp_x[ind - 1] <= b_tec.var_hp_p_el[t]

                dis.const_hp_on1 = pyo.Constraint(rule=init_hp_on1)

                def init_hp_on2(const):
                    return b_tec.var_hp_p_el[t] <= st_hp_bp_x[ind]

                dis.const_hp_on2 = pyo.Constraint(rule=init_hp_on2)

                # HP performance
                def init_hp_performance(const):
                    if self.component_options.other["component"] == "DB":
                        add = st_hp_alpha_db[t - 1, ind - 1] * b_tec.var_db_input[t]
                    elif self.component_options.other["component"] == "OHB":
                        add = (
                            st_hp_alpha_ohb[t - 1, ind - 1] * b_tec.var_ohb_h2_input[t]
                        )
                    else:
                        add = 0

                    return (
                        b_tec.var_hp_p_el[t]
                        == st_hp_alpha_gt[t - 1, ind - 1] * b_tec.var_gt_p_th[t]
                        + st_hp_alpha_hp[t - 1, ind - 1] * b_tec.var_hp_p_hp[t]
                        + st_hp_alpha_mp[t - 1, ind - 1] * b_tec.var_hp_p_mp[t]
                        + add
                        + st_hp_alpha_cst[t - 1, ind - 1]
                    )

                dis.const_hp_performance = pyo.Constraint(rule=init_hp_performance)

        b_tec.dis_performance_hp = gdp.Disjunct(
            self.set_t_performance, s_indicators, rule=init_perf_hp
        )

        # Bind disjuncts
        def bind_disjunctions(dis, t):
            return [b_tec.dis_performance_hp[t, i] for i in s_indicators]

        b_tec.disjunction_performance_hp = gdp.Disjunction(
            self.set_t_performance, rule=bind_disjunctions
        )

        st_mp_alpha_gt = coeff_ti["MP"]["alpha_gt"]
        st_mp_alpha_hp = coeff_ti["MP"]["alpha_hp"]
        st_mp_alpha_mp = coeff_ti["MP"]["alpha_mp"]
        st_mp_alpha_cst = coeff_ti["MP"]["alpha_cst"]
        st_mp_bp_x = coeff_ti["MP"]["bp_mp"]

        if self.component_options.other["component"] == "DB":
            st_mp_alpha_db = coeff_ti["MP"]["alpha_db"]
        elif self.component_options.other["component"] == "OHB":
            st_mp_alpha_ohb = coeff_ti["MP"]["alpha_ohb"]

        s_indicators = range(0, nr_segments + 1)

        def init_perf_mp(dis, t, ind):
            if ind == 0:  # technology off
                dis.const_x_off = pyo.Constraint(expr=b_tec.var_x[t] == 0)

            else:  # technology on
                dis.const_x_on = pyo.Constraint(expr=b_tec.var_x[t] == 1)

                # MP bounds
                def init_mp_on1(const):
                    return st_mp_bp_x[ind - 1] <= b_tec.var_mp_p_el[t]

                dis.const_mp_on1 = pyo.Constraint(rule=init_mp_on1)

                def init_mp_on2(const):
                    return b_tec.var_mp_p_el[t] <= st_mp_bp_x[ind]

                dis.const_mp_on2 = pyo.Constraint(rule=init_mp_on2)

                # MP performance
                def init_mp_performance(const):
                    if self.component_options.other["component"] == "DB":
                        add = st_mp_alpha_db[t - 1, ind - 1] * b_tec.var_db_input[t]
                    elif self.component_options.other["component"] == "OHB":
                        add = (
                            st_mp_alpha_ohb[t - 1, ind - 1] * b_tec.var_ohb_h2_input[t]
                        )
                    else:
                        add = 0
                    return (
                        b_tec.var_mp_p_el[t]
                        == st_mp_alpha_gt[t - 1, ind - 1] * b_tec.var_gt_p_th[t]
                        + st_mp_alpha_hp[t - 1, ind - 1] * b_tec.var_mp_p_hp[t]
                        + st_mp_alpha_mp[t - 1, ind - 1] * b_tec.var_mp_p_mp[t]
                        + add
                        + st_mp_alpha_cst[t - 1, ind - 1]
                    )

                dis.const_mp_performance = pyo.Constraint(rule=init_mp_performance)

        b_tec.dis_performance_mp = gdp.Disjunct(
            self.set_t_performance, s_indicators, rule=init_perf_mp
        )

        # Bind disjuncts
        def bind_disjunctions(dis, t):
            return [b_tec.dis_performance_mp[t, i] for i in s_indicators]

        b_tec.disjunction_performance_mp = gdp.Disjunction(
            self.set_t_performance, rule=bind_disjunctions
        )

        return b_tec

    def write_results_tec_operation(self, h5_group, model_block):
        """
        Function to report results of technologies after optimization

        :param b_tec: technology model block
        :return: dict results: holds results
        """
        super(CCPP, self).write_results_tec_operation(h5_group, model_block)

        h5_group.create_dataset(
            "gt_input",
            data=[model_block.var_gt_input[t].value for t in self.set_t_performance],
        )
        h5_group.create_dataset(
            "gt_ng_input",
            data=[model_block.var_gt_ng_input[t].value for t in self.set_t_performance],
        )
        h5_group.create_dataset(
            "gt_h2_input",
            data=[model_block.var_gt_h2_input[t].value for t in self.set_t_performance],
        )
        h5_group.create_dataset(
            "gt_p_el",
            data=[model_block.var_gt_p_el[t].value for t in self.set_t_performance],
        )
        h5_group.create_dataset(
            "gt_p_th",
            data=[model_block.var_gt_p_th[t].value for t in self.set_t_performance],
        )
        h5_group.create_dataset(
            "hp_p_el",
            data=[model_block.var_hp_p_el[t].value for t in self.set_t_performance],
        )
        h5_group.create_dataset(
            "mp_p_el",
            data=[model_block.var_mp_p_el[t].value for t in self.set_t_performance],
        )
        if self.component_options.other["component"] == "DB":
            h5_group.create_dataset(
                "db_input",
                data=[
                    model_block.var_db_input[t].value for t in self.set_t_performance
                ],
            )
            h5_group.create_dataset(
                "db_h2_input",
                data=[
                    model_block.var_db_h2_input[t].value for t in self.set_t_performance
                ],
            )
            h5_group.create_dataset(
                "db_ng_input",
                data=[
                    model_block.var_db_ng_input[t].value for t in self.set_t_performance
                ],
            )
        if self.component_options.other["component"] == "OHB":
            h5_group.create_dataset(
                "ohb_h2_input",
                data=[
                    model_block.var_ohb_h2_input[t].value
                    for t in self.set_t_performance
                ],
            )

    def _define_ramping_rates(self, b_tec, data):
        """
        Constraints the inputs for a ramping rate

        :param b_tec: technology model block
        :return:
        """
        dynamics = self.processed_coeff.dynamics

        ramping_time = dynamics["ramping_time"]

        # Calculate ramping rates
        if "ref_size" in dynamics and not dynamics["ref_size"] == -1:
            ramping_rate = dynamics["ref_size"] / ramping_time
        else:
            ramping_rate = b_tec.var_size / ramping_time

        # Constraints ramping rates
        if "ramping_const_int" in dynamics and dynamics["ramping_const_int"] == 1:

            s_indicators = range(0, 3)

            def init_ramping_operation_on(dis, t, ind):
                if t > 1:
                    if ind == 0:  # ramping constrained
                        dis.const_ramping_on = pyo.Constraint(
                            expr=b_tec.var_x[t] - b_tec.var_x[t - 1] == 0
                        )

                        def init_ramping_down_rate_operation(const):
                            return -ramping_rate <= sum(
                                self.input[t, car_input] - self.input[t - 1, car_input]
                                for car_input in b_tec.set_input_carriers
                            )

                        dis.const_ramping_down_rate = pyo.Constraint(
                            rule=init_ramping_down_rate_operation
                        )

                        def init_ramping_up_rate_operation(const):
                            return (
                                sum(
                                    self.input[t, car_input]
                                    - self.input[t - 1, car_input]
                                    for car_input in b_tec.set_input_carriers
                                )
                                <= ramping_rate
                            )

                        dis.const_ramping_up_rate = pyo.Constraint(
                            rule=init_ramping_up_rate_operation
                        )

                    elif ind == 1:  # startup, no ramping constraint
                        dis.const_ramping_on = pyo.Constraint(
                            expr=b_tec.var_x[t] - b_tec.var_x[t - 1] == 1
                        )

                    else:  # shutdown, no ramping constraint
                        dis.const_ramping_on = pyo.Constraint(
                            expr=b_tec.var_x[t] - b_tec.var_x[t - 1] == -1
                        )

            b_tec.dis_ramping_operation_on = gdp.Disjunct(
                self.set_t_performance, s_indicators, rule=init_ramping_operation_on
            )

            # Bind disjuncts
            def bind_disjunctions(dis, t):
                return [b_tec.dis_ramping_operation_on[t, i] for i in s_indicators]

            b_tec.disjunction_ramping_operation_on = gdp.Disjunction(
                self.set_t_performance, rule=bind_disjunctions
            )

        else:
            if data["config"]["optimization"]["typicaldays"]["N"]["value"] == 0:
                input_aux_rr = self.input
                set_t_rr = self.set_t_performance
            else:
                if (
                    data["config"]["optimization"]["typicaldays"]["method"]["value"]
                    == 1
                ):
                    sequence = data["k_means_specs"]["sequence"]
                elif (
                    data["config"]["optimization"]["typicaldays"]["method"]["value"]
                    == 2
                ):
                    sequence = self.sequence

                # init bounds at full res
                bounds_rr_full = {
                    "input": self.fitting_class.calculate_input_bounds(
                        self.component_options.size_based_on, len(self.set_t_full)
                    )
                }

                # create input variable for full res
                def init_input_bounds(bounds, t, car):
                    return tuple(
                        bounds_rr_full["input"][car][t - 1, :]
                        * self.processed_coeff.time_independent["size_max"]
                        * self.processed_coeff.time_independent["rated_power"]
                    )

                b_tec.var_input_rr_full = pyo.Var(
                    self.set_t_full,
                    b_tec.set_input_carriers,
                    within=pyo.NonNegativeReals,
                    bounds=init_input_bounds,
                )

                b_tec.const_link_full_resolution_rr = link_full_resolution_to_clustered(
                    self.input,
                    b_tec.var_input_rr_full,
                    self.set_t_full,
                    sequence,
                    b_tec.set_input_carriers,
                )

                input_aux_rr = b_tec.var_input_rr_full
                set_t_rr = self.set_t_full

            # Ramping constraint without integers
            def init_ramping_down_rate(const, t):
                if t > 1:
                    return -ramping_rate <= sum(
                        input_aux_rr[t, car_input] - input_aux_rr[t - 1, car_input]
                        for car_input in b_tec.set_input_carriers
                    )
                else:
                    return pyo.Constraint.Skip

            b_tec.const_ramping_down_rate = pyo.Constraint(
                set_t_rr, rule=init_ramping_down_rate
            )

            def init_ramping_up_rate(const, t):
                if t > 1:
                    return (
                        sum(
                            input_aux_rr[t, car_input] - input_aux_rr[t - 1, car_input]
                            for car_input in b_tec.set_input_carriers
                        )
                        <= ramping_rate
                    )
                else:
                    return pyo.Constraint.Skip

            b_tec.const_ramping_up_rate = pyo.Constraint(
                set_t_rr, rule=init_ramping_up_rate
            )

        return b_tec



import statsmodels.api as sm
import numpy as np

from ..utilities import (
    fit_linear_function,
    fit_piecewise_function,
    sig_figs,
)


class FitGenericTecTypeType1:
    """
    Class to fit performance of type1 performance functions (linear, through origin)
    out = alpha1 * in
    """

    def __init__(self, params):
        self.input_parameters = params
        self.coeff = {}
        self.bounds = {}

    def fit_performance_function(self, performance_data: dict):
        """
        Fits performance function for input-output data for type 1 technologies

        :param dict performance_data: performance data
        """
        x = performance_data["in"]

        for car in performance_data["out"]:
            self.coeff[car] = {}
            y = performance_data["out"][car]
            fit = fit_linear_function(x, y)
            self.coeff[car]["alpha1"] = sig_figs(fit[0], 6)

        return self.coeff

    def calculate_input_bounds(self, size_based_on: str, time_steps: int):
        """
        Calculates input bounds for type 1 generic technologies

        :param str size_based_on: 'input' or 'output'
        :param int time_steps: number of time steps
        """
        input_bounds = {}

        if size_based_on == "input":
            for car in self.input_parameters.input_carrier:
                input_bounds[car] = np.column_stack(
                    (np.zeros(shape=(time_steps)), np.ones(shape=(time_steps)))
                )
        elif size_based_on == "output":
            for car in self.input_parameters.input_carrier:
                if car in self.coeff:
                    car_aux = car
                else:
                    car_aux = "out"
                input_bounds[car] = np.column_stack(
                    (
                        np.zeros(shape=(time_steps)),
                        np.ones(shape=(time_steps)) / self.coeff[car_aux]["alpha1"],
                    )
                )
        else:
            raise Exception("size_based_on must be either input or output")

        return input_bounds

    def calculate_output_bounds(self, size_based_on: str, time_steps: int):
        """
        Calculates output bounds for type 1 generic technologies

        :param str size_based_on: 'input' or 'output'
        :param int time_steps: number of time steps
        """
        output_bounds = {}

        if size_based_on == "input":
            for car in self.input_parameters.output_carrier:
                if car in self.coeff:
                    car_aux = car
                else:
                    car_aux = "out"
                output_bounds[car] = np.column_stack(
                    (
                        np.zeros(shape=(time_steps)),
                        np.ones(shape=(time_steps)) * self.coeff[car_aux]["alpha1"],
                    )
                )
        elif size_based_on == "output":
            for car in self.input_parameters.output_carrier:
                output_bounds[car] = np.column_stack(
                    (np.zeros(shape=(time_steps)), np.ones(shape=(time_steps)))
                )
        else:
            raise Exception("size_based_on must be either input or output")

        return output_bounds


class FitGenericTecTypeType2:
    """
    Class to fit performance of type1 performance functions (linear, with min partload)
    out = alpha1 * in + alpha2
    (out - alpha2)/alpha1
    """

    def __init__(self, params):
        self.input_parameters = params
        self.coeff = {}
        self.bounds = {}

    def fit_performance_function(self, performance_data: dict):
        """
        Fits performance function for input-output data for type 2 technologies

        :param dict performance_data: performance data
        """
        x = performance_data["in"]
        x = sm.add_constant(x)

        for car in performance_data["out"]:
            self.coeff[car] = {}
            y = performance_data["out"][car]
            fit = fit_linear_function(x, y)
            self.coeff[car]["alpha1"] = sig_figs(fit[1], 6)
            self.coeff[car]["alpha2"] = sig_figs(fit[0], 6)

        return self.coeff

    def calculate_input_bounds(self, size_based_on: str, time_steps: int):
        """
        Calculates input bounds for type 2 generic technologies

        :param str size_based_on: 'input' or 'output'
        :param int time_steps: number of time steps
        """
        input_bounds = {}

        if size_based_on == "input":
            for car in self.input_parameters.input_carrier:
                input_bounds[car] = np.column_stack(
                    (np.zeros(shape=time_steps), np.ones(shape=time_steps))
                )
        elif size_based_on == "output":
            for car in self.input_parameters.input_carrier:
                if car in self.coeff:
                    car_aux = car
                else:
                    car_aux = "out"
                input_bounds[car] = (
                    np.column_stack(
                        (np.zeros(shape=time_steps), np.ones(shape=time_steps))
                    )
                    - self.coeff[car_aux]["alpha2"]
                ) / self.coeff[car_aux]["alpha1"]
        else:
            raise Exception("size_based_on must be either input or output")

        return input_bounds

    def calculate_output_bounds(self, size_based_on: str, time_steps: int):
        """
        Calculates output bounds for type 2 generic technologies

        :param str size_based_on: 'input' or 'output'
        :param int time_steps: number of time steps
        """
        output_bounds = {}

        if size_based_on == "input":
            for car in self.input_parameters.output_carrier:
                if car in self.coeff:
                    car_aux = car
                else:
                    car_aux = "out"
                output_bounds[car] = np.column_stack(
                    (
                        np.zeros(shape=(time_steps)),
                        np.ones(shape=(time_steps)) * self.coeff[car_aux]["alpha1"]
                        + self.coeff[car_aux]["alpha2"],
                    )
                )
        elif size_based_on == "output":
            for car in self.input_parameters.output_carrier:
                output_bounds[car] = np.column_stack(
                    (np.zeros(shape=time_steps), np.ones(shape=time_steps))
                )
        else:
            raise Exception("size_based_on must be either input or output")

        return output_bounds


class FitGenericTecTypeType34:
    """
    Class to fit performance of type3 performance functions (piecewise linear, with min partload)
    out = alpha1[i] * in + alpha2
    """

    def __init__(self, params):
        self.input_parameters = params
        self.coeff = {}
        self.bounds = {}

    def fit_performance_function(self, performance_data: dict):
        """
        Fits performance function for input-output data for type 3/4 technologies

        :param dict performance_data: performance data
        """
        if "nr_segments_piecewise" in performance_data:
            nr_seg = performance_data["nr_segments_piecewise"]
        else:
            nr_seg = 2
        x = performance_data["in"]
        y = performance_data["out"]
        self.coeff = fit_piecewise_function(x, y, nr_seg)

        return self.coeff

    def calculate_input_bounds(self, size_based_on: str, time_steps: int):
        """
        Calculates input bounds for type 3/4 generic technologies

        :param str size_based_on: 'input' or 'output'
        :param int time_steps: number of time steps
        """
        input_bounds = {}

        if size_based_on == "input":
            for car in self.input_parameters.input_carrier:
                input_bounds[car] = np.column_stack(
                    (np.zeros(shape=(time_steps)), np.ones(shape=(time_steps)))
                )
        elif size_based_on == "output":
            for car in self.input_parameters.input_carrier:
                if car in self.coeff:
                    car_aux = car
                else:
                    car_aux = "out"
                input_bounds[car] = (
                    np.column_stack(
                        (np.zeros(shape=time_steps), np.ones(shape=time_steps))
                    )
                    - self.coeff[car_aux]["alpha2"][-1]
                ) / self.coeff[car_aux]["alpha1"][-1]
        else:
            raise Exception("size_based_on must be either input or output")

        return input_bounds

    def calculate_output_bounds(self, size_based_on: str, time_steps: int):
        """
        Calculates output bounds for type 3/4 generic technologies

        :param str size_based_on: 'input' or 'output'
        :param int time_steps: number of time steps
        """
        output_bounds = {}

        if size_based_on == "input":
            for car in self.input_parameters.output_carrier:
                if car in self.coeff:
                    car_aux = car
                else:
                    car_aux = "out"
                output_bounds[car] = np.column_stack(
                    (
                        np.zeros(shape=(time_steps)),
                        np.ones(shape=(time_steps)) * self.coeff[car_aux]["alpha1"][-1]
                        + self.coeff[car_aux]["alpha2"][-1],
                    )
                )
        elif size_based_on == "output":
            for car in self.input_parameters.output_carrier:
                output_bounds[car] = np.column_stack(
                    (np.zeros(shape=(time_steps)), np.ones(shape=(time_steps)))
                )
        else:
            raise Exception("size_based_on must be either input or output")

        return output_bounds



import pyomo.environ as pyo
import pyomo.gdp as gdp
import copy
import pandas as pd
import numpy as np
from pathlib import Path
from scipy.interpolate import griddata

from ..utilities import fit_piecewise_function
from ..technology import Technology

import logging

log = logging.getLogger(__name__)


class DacAdsorption(Technology):
    """
    Direct Air Capture technology (adsorption)

    The model resembles as Direct Air Capture technology with a modular setup. It
    has a heat and electricity input and CO2 as an output. The performance is based
    on data for a generic solid sorbent, as described in the article (see below).
    The performance data is fitted to the ambient temperature and humidity at the
    respective node.

    The model is based on Wiegner et al. (2022). Optimal Design and Operation of
    Solid Sorbent Direct Air Capture Processes at Varying Ambient Conditions.
    Industrial and Engineering Chemistry Research, 2022, 12649–12667.
    https://doi.org/10.1021/acs.iecr.2c00681. It resembles operation configuration 1
    without water spraying.
    """

    def __init__(self, tec_data: dict):
        """
        Constructor

        :param dict tec_data: technology data
        """
        super().__init__(tec_data)

        self.component_options.emissions_based_on = "output"
        self.component_options.main_output_carrier = "CO2captured"

    def fit_technology_performance(self, climate_data: pd.DataFrame, location: dict):
        """
        Fits the technology performance

        :param pd.Dataframe climate_data: dataframe containing climate data
        :param dict location: dict containing location details
        """
        super(DacAdsorption, self).fit_technology_performance(climate_data, location)

        # Number of segments
        nr_segments = self.input_parameters.performance_data["nr_segments"]

        # Read performance data from file
        performance_data_path = Path(__file__).parent.parent.parent.parent
        performance_data_path = (
            performance_data_path
            / "data/technology_data/CO2Capture/DAC_adsorption_data/dac_adsorption_performance.txt"
        )

        performance_data = pd.read_csv(performance_data_path, sep=",")
        performance_data = performance_data.rename(
            columns={"T": "temp_air", "RH": "humidity"}
        )

        # Unit Conversion of input data
        performance_data.E_tot = performance_data.E_tot.multiply(
            performance_data.CO2_Out / 3600
        )  # in MWh / h
        performance_data.E_el = performance_data.E_el.multiply(
            performance_data.CO2_Out / 3600
        )  # in MWh / h
        performance_data.E_th = performance_data.E_th.multiply(
            performance_data.CO2_Out / 3600
        )  # in MWh / h
        performance_data.CO2_Out = performance_data.CO2_Out / 1000  # in t / h

        # Get humidity and temperature
        RH = copy.deepcopy(climate_data["rh"])
        T = copy.deepcopy(climate_data["temp_air"])

        # Set minimum temperature
        T.loc[T < min(performance_data.temp_air)] = min(performance_data.temp_air)

        # Derive performance points for each timestep
        def interpolate_performance_point(t, rh, point_data, var):
            zi = griddata(
                (point_data.temp_air, point_data.humidity),
                point_data[var],
                (T, RH),
                method="linear",
            )
            return zi

        CO2_Out = np.empty(shape=(len(T), len(performance_data.Point.unique())))
        E_tot = np.empty(shape=(len(T), len(performance_data.Point.unique())))
        E_el = np.empty(shape=(len(T), len(performance_data.Point.unique())))
        for point in performance_data.Point.unique():
            CO2_Out[:, point - 1] = interpolate_performance_point(
                T, RH, performance_data.loc[performance_data.Point == point], "CO2_Out"
            )
            E_tot[:, point - 1] = interpolate_performance_point(
                T, RH, performance_data.loc[performance_data.Point == point], "E_tot"
            )
            E_el[:, point - 1] = interpolate_performance_point(
                T, RH, performance_data.loc[performance_data.Point == point], "E_el"
            )

        # Derive piecewise definition
        alpha = np.empty(shape=(len(T), nr_segments))
        beta = np.empty(shape=(len(T), nr_segments))
        b = np.empty(shape=(len(T), nr_segments + 1))
        gamma = np.empty(shape=(len(T), nr_segments))
        delta = np.empty(shape=(len(T), nr_segments))
        a = np.empty(shape=(len(T), nr_segments + 1))
        el_in_max = np.empty(shape=(len(T)))
        th_in_max = np.empty(shape=(len(T)))
        out_max = np.empty(shape=(len(T)))
        total_in_max = np.empty(shape=(len(T)))

        log.info("Deriving performance data for DAC...")

        for timestep in range(len(T)):
            if timestep % 100 == 1:
                print("\rComplete: ", round(timestep / len(T), 2) * 100, "%", end="")
            # Input-Output relation
            y = {}
            y["CO2_Out"] = CO2_Out[timestep, :]
            time_step_fit = fit_piecewise_function(
                E_tot[timestep, :], y, int(nr_segments)
            )
            alpha[timestep, :] = time_step_fit["CO2_Out"]["alpha1"]
            beta[timestep, :] = time_step_fit["CO2_Out"]["alpha2"]
            b[timestep, :] = time_step_fit["CO2_Out"]["bp_x"]
            out_max[timestep] = max(time_step_fit["CO2_Out"]["bp_y"])
            total_in_max[timestep] = max(time_step_fit["CO2_Out"]["bp_x"])

            # Input-Input relation
            y = {}
            y["E_el"] = E_el[timestep, :]
            time_step_fit = fit_piecewise_function(
                E_tot[timestep, :], y, int(nr_segments)
            )
            gamma[timestep, :] = time_step_fit["E_el"]["alpha1"]
            delta[timestep, :] = time_step_fit["E_el"]["alpha2"]
            a[timestep, :] = time_step_fit["E_el"]["bp_x"]
            el_in_max[timestep] = max(time_step_fit["E_el"]["bp_y"])
            th_in_max[timestep] = max(time_step_fit["E_el"]["bp_x"])

        print("Complete: ", 100, "%")

        # Coefficients
        self.processed_coeff.time_dependent_full["alpha"] = alpha
        self.processed_coeff.time_dependent_full["beta"] = beta
        self.processed_coeff.time_dependent_full["b"] = b
        self.processed_coeff.time_dependent_full["gamma"] = gamma
        self.processed_coeff.time_dependent_full["delta"] = delta
        self.processed_coeff.time_dependent_full["a"] = a
        self.processed_coeff.time_dependent_full["out_max"] = out_max
        self.processed_coeff.time_dependent_full["el_in_max"] = el_in_max
        self.processed_coeff.time_dependent_full["th_in_max"] = th_in_max
        self.processed_coeff.time_dependent_full["total_in_max"] = total_in_max

        self.processed_coeff.time_independent["eta_elth"] = (
            self.input_parameters.performance_data["performance"]["eta_elth"]
        )

        # Options
        self.component_options.other["nr_segments"] = (
            self.input_parameters.performance_data["nr_segments"]
        )
        self.component_options.other["ohmic_heating"] = (
            self.input_parameters.performance_data["ohmic_heating"]
        )

    def _calculate_bounds(self):
        """
        Calculates the bounds of the variables used
        """
        super(DacAdsorption, self)._calculate_bounds()

        time_steps = len(self.set_t_performance)

        # Output Bounds
        self.bounds["output"]["CO2captured"] = np.column_stack(
            (
                np.zeros(shape=(time_steps)),
                self.processed_coeff.time_dependent_used["out_max"],
            )
        )

        # Input Bounds
        self.bounds["input"]["electricity"] = np.column_stack(
            (
                np.zeros(shape=(time_steps)),
                self.processed_coeff.time_dependent_used["el_in_max"]
                + self.processed_coeff.time_dependent_used["th_in_max"]
                / self.input_parameters.performance_data["performance"]["eta_elth"],
            )
        )
        self.bounds["input"]["heat"] = np.column_stack(
            (
                np.zeros(shape=(time_steps)),
                self.processed_coeff.time_dependent_used["th_in_max"],
            )
        )
        self.bounds["input"]["total"] = np.column_stack(
            (
                np.zeros(shape=(time_steps)),
                self.processed_coeff.time_dependent_used["total_in_max"],
            )
        )

    def construct_tech_model(self, b_tec, data: dict, set_t_full, set_t_clustered):
        """
        Adds constraints to technology blocks for tec_type DAC_adsorption

        :param b_tec: pyomo block with technology model
        :param dict data: data containing model configuration
        :param set_t_full: pyomo set containing timesteps
        :param set_t_clustered: pyomo set containing clustered timesteps
        :return: pyomo block with technology model
        """
        super(DacAdsorption, self).construct_tech_model(
            b_tec, data, set_t_full, set_t_clustered
        )

        # Comments on the equations refer to the equation numbers in the paper. All equations can be looked up there.

        # Transformation required
        self.big_m_transformation_required = 1

        # DATA OF TECHNOLOGY
        nr_segments = self.component_options.other["nr_segments"]
        ohmic_heating = self.component_options.other["ohmic_heating"]

        bounds = self.bounds
        coeff_td = self.processed_coeff.time_dependent_used
        coeff_ti = self.processed_coeff.time_independent

        alpha = coeff_td["alpha"]
        beta = coeff_td["beta"]
        b_point = coeff_td["b"]
        gamma = coeff_td["gamma"]
        delta = coeff_td["delta"]
        a_point = coeff_td["a"]
        eta_elth = coeff_ti["eta_elth"]

        # Additional sets
        b_tec.set_pieces = pyo.RangeSet(1, nr_segments)

        # Additional decision variables
        b_tec.var_modules_on = pyo.Var(
            self.set_t_performance,
            domain=pyo.NonNegativeIntegers,
            bounds=(b_tec.para_size_min, b_tec.para_size_max),
        )

        def init_input_total_bounds(bds, t):
            return tuple(bounds["input"]["total"][t - 1] * b_tec.para_size_max)

        b_tec.var_input_total = pyo.Var(
            self.set_t_performance,
            within=pyo.NonNegativeReals,
            bounds=init_input_total_bounds,
        )

        def init_input_el_bounds(bds, t):
            return tuple(bounds["input"]["electricity"][t - 1] * b_tec.para_size_max)

        b_tec.var_input_el = pyo.Var(
            self.set_t_performance,
            within=pyo.NonNegativeReals,
            bounds=init_input_el_bounds,
        )

        def init_input_th_bounds(bds, t):
            return tuple(bounds["input"]["heat"][t - 1] * b_tec.para_size_max)

        b_tec.var_input_th = pyo.Var(
            self.set_t_performance,
            within=pyo.NonNegativeReals,
            bounds=init_input_th_bounds,
        )

        def init_input_ohmic_bounds(bds, t):
            return tuple(
                (bounds["input"]["heat"][t - 1] / eta_elth * b_tec.para_size_max)
            )

        b_tec.var_input_ohmic = pyo.Var(
            self.set_t_performance,
            within=pyo.NonNegativeReals,
            bounds=init_input_ohmic_bounds,
        )

        # Input-Output relationship (eq. 1-5)
        def init_input_output(dis, t, ind):
            # Input-output (eq. 2)
            def init_output(const):
                return (
                    self.output[t, "CO2captured"]
                    == alpha[t - 1, ind - 1] * b_tec.var_input_total[t]
                    + beta[t - 1, ind - 1] * b_tec.var_modules_on[t]
                )

            dis.const_output = pyo.Constraint(rule=init_output)

            # Lower bound on the energy input (eq. 5)
            def init_input_low_bound(const):
                return (
                    b_point[t - 1, ind - 1] * b_tec.var_modules_on[t]
                    <= b_tec.var_input_total[t]
                )

            dis.const_input_on1 = pyo.Constraint(rule=init_input_low_bound)

            # Upper bound on the energy input (eq. 5)
            def init_input_up_bound(const):
                return (
                    b_tec.var_input_total[t]
                    <= b_point[t - 1, ind] * b_tec.var_modules_on[t]
                )

            dis.const_input_on2 = pyo.Constraint(rule=init_input_up_bound)

        b_tec.dis_input_output = gdp.Disjunct(
            self.set_t_performance, b_tec.set_pieces, rule=init_input_output
        )

        # Bind disjuncts
        def bind_disjunctions(dis, t):
            return [b_tec.dis_input_output[t, i] for i in b_tec.set_pieces]

        b_tec.disjunction_input_output = gdp.Disjunction(
            self.set_t_performance, rule=bind_disjunctions
        )

        # Electricity-Heat relationship (eq. 7-10)
        def init_input_input(dis, t, ind):
            # Input-output (eq. 7)
            def init_input(const):
                return (
                    b_tec.var_input_el[t]
                    == gamma[t - 1, ind - 1] * b_tec.var_input_total[t]
                    + delta[t - 1, ind - 1] * b_tec.var_modules_on[t]
                )

            dis.const_output = pyo.Constraint(rule=init_input)

            # Lower bound on the energy input (eq. 10)
            def init_input_low_bound(const):
                return (
                    a_point[t - 1, ind - 1] * b_tec.var_modules_on[t]
                    <= b_tec.var_input_total[t]
                )

            dis.const_input_on1 = pyo.Constraint(rule=init_input_low_bound)

            # Upper bound on the energy input (eq. 10)
            def init_input_up_bound(const):
                return (
                    b_tec.var_input_total[t]
                    <= a_point[t - 1, ind] * b_tec.var_modules_on[t]
                )

            dis.const_input_on2 = pyo.Constraint(rule=init_input_up_bound)

        b_tec.dis_input_input = gdp.Disjunct(
            self.set_t_performance, b_tec.set_pieces, rule=init_input_input
        )

        # Bind disjuncts
        def bind_disjunctions(dis, t):
            return [b_tec.dis_input_input[t, i] for i in b_tec.set_pieces]

        b_tec.disjunction_input_input = gdp.Disjunction(
            self.set_t_performance, rule=bind_disjunctions
        )

        # Constraint of number of working modules (eq. 6)
        def init_modules_on(const, t):
            return b_tec.var_modules_on[t] <= b_tec.var_size

        b_tec.const_var_modules_on = pyo.Constraint(
            self.set_t_performance, rule=init_modules_on
        )

        # Connection thermal and electric energy demand (eq. 11)
        def init_thermal_energy(const, t):
            return (
                b_tec.var_input_th[t]
                == b_tec.var_input_total[t] - b_tec.var_input_el[t]
            )

        b_tec.const_thermal_energy = pyo.Constraint(
            self.set_t_performance, rule=init_thermal_energy
        )

        # Account for ohmic heating (eq. 12)
        def init_input_el(const, t):
            return (
                self.input[t, "electricity"]
                == b_tec.var_input_ohmic[t] + b_tec.var_input_el[t]
            )

        b_tec.const_input_el = pyo.Constraint(
            self.set_t_performance, rule=init_input_el
        )

        def init_input_th(const, t):
            return (
                self.input[t, "heat"]
                == b_tec.var_input_th[t] - b_tec.var_input_ohmic[t] * eta_elth
            )

        b_tec.const_input_th = pyo.Constraint(
            self.set_t_performance, rule=init_input_th
        )

        # If ohmic heating not allowed, set to zero
        if not ohmic_heating:

            def init_ohmic_heating(const, t):
                return b_tec.var_input_ohmic[t] == 0

            b_tec.const_ohmic_heating = pyo.Constraint(
                self.set_t_performance, rule=init_ohmic_heating
            )

        return b_tec

    def write_results_tec_operation(self, h5_group, model_block):
        """
        Function to report technology operation

        :param model_block: pyomo network block
        :param h5_group: h5 group to write to
        """
        super(DacAdsorption, self).write_results_tec_operation(h5_group, model_block)

        h5_group.create_dataset(
            "modules_on",
            data=[
                model_block.var_modules_on[self.sequence[t - 1]].value
                for t in self.set_t_performance
            ],
        )
        h5_group.create_dataset(
            "ohmic_heating",
            data=[
                model_block.var_input_ohmic[self.sequence[t - 1]].value
                for t in self.set_t_performance
            ],
        )
